# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__


from abc import ABC, abstractmethod

import pandas as pd
import numpy as np

from pylife import PylifeSignal

from .abstract_load_collective import AbstractLoadCollective


@pd.api.extensions.register_series_accessor('load_collective')
class LoadHistogram(PylifeSignal, AbstractLoadCollective):

    def _validate(self):
        self._class_location = 'mid'
        if 'range' in self._obj.index.names:
            self._fail_if_not_multiindex(['range', 'mean'])
            self._impl = _RangeMeanMatrix(self._obj)
            return
        if 'from' in self._obj.index.names and 'to' in self._obj.index.names:
            self._fail_if_not_multiindex(['from', 'to'])
            self._impl = _FromToMatrix(self._obj)
            return

        raise AttributeError("Load collective matrix needs either 'range'/('mean') or 'from'/'to' in index levels.")

    def _fail_if_not_multiindex(self, index_names):
        for name in index_names:
            if name not in self._obj.index.names:
                continue
            if not isinstance(self._obj.index.get_level_values(name), pd.IntervalIndex):
                raise AttributeError("Index of a load collective matrix must be pandas.IntervalIndex.")

    @property
    def amplitude(self):
        """Calculate the amplitudes of the load collective.

        Returns
        -------
        amplitude : pd.Series
            The amplitudes of the load collective
        """
        rng = self._impl.amplitude()
        return pd.Series(rng/2., name='amplitude', index=self._obj.index)

    @property
    def amplitude_histogram(self):
        index = self._impl.amplitude_histogram_index()
        index.name = 'amplitude'
        return pd.Series(self._obj.values, index=index, name='cycles')

    @property
    def meanstress(self):
        """Calculate the mean load values of the load collective.

        Returns
        -------
        mean : pd.Series
            The mean load values of the load collective
        """
        mean = self._impl.meanstress()
        return pd.Series(mean, name='meanstress', index=self._obj.index)

    @property
    def R(self):
        """Calculate the R values of the load collective.

        Returns
        -------
        R : pd.Series
            The R values of the load collective
        """
        res = (self.lower / self.upper).fillna(0.0)
        res.name = 'R'
        return res

    @property
    def upper(self):
        """Calculate the upper load values of the load collective.

        Returns
        -------
        upper : pd.Series
            The upper load values of the load collective
        """
        res = self.meanstress + self.amplitude
        res.name = 'upper'
        return res

    @property
    def lower(self):
        """Calculate the lower load values of the load collective.

        Returns
        -------
        lower : pd.Series
            The lower load values of the load collective
        """
        res = self.meanstress - self.amplitude
        res.name = 'lower'
        return res

    @property
    def cycles(self):
        """The cycles of each class of the collective.

        Returns
        -------
        cycles : pd.Series
            The cycles of each class of the collective
        """
        cycles = self._obj.copy()
        cycles.name = 'cycles'
        return cycles

    def use_class_right(self):
        """Use the upper limit of the class bins.


        Returns
        -------
        self
        """
        self._impl._class_location = 'right'
        return self

    def use_class_left(self):
        """Use the lower limit of the class bins.

        Returns
        -------
        self
        """
        self._impl._class_location = 'left'
        return self

    def scale(self, factors):
        """Scale the collective.

        Parameters
        ----------
        factors : scalar or :class:`pandas.Series`
            The factor(s) to scale the collective with.

        Returns
        -------
        scaled : ``LoadCollective``
            The scaled collective.
        """
        return self._shift_or_scale(lambda x, y: x * y, factors).load_collective

    def shift(self, diffs):
        """Shift the collective.

        Parameters
        ----------
        diffs : scalar or :class:`pandas.Series`
            The diff(s) to shift the collective by.

        Returns
        -------
        shifted : ``LoadCollective``
            The shifted collective.
        """
        return self._shift_or_scale(lambda x, y: x + y, diffs, skip=['range']).load_collective

    def _shift_or_scale(self, func, operand, skip=None):
        def do_transform_interval_index(level_name):
            level = obj.index.get_level_values(level_name)
            if level.name not in self._impl.index_names or level_name in skip:
                return level
            values = level.values
            left = func(values.left, operand_broadcast)
            right = func(values.right, operand_broadcast)

            index = pd.IntervalIndex.from_arrays(left, right)
            return index

        skip = skip or []
        operand_broadcast, obj = self.broadcast(operand)

        levels = [do_transform_interval_index(lv) for lv in obj.index.names]

        new_index = pd.MultiIndex.from_arrays(levels, names=obj.index.names)
        return pd.Series(obj.values, index=new_index, name='cycles')

    def cumulated_range(self):
        return pd.Series(self._obj.groupby('range').transform(lambda g: np.cumsum(g)),
                         name='cumulated_cycles')

class _LoadHistogramImpl(ABC):

    @property
    @abstractmethod
    def index_names(self):
        return set([])

    def __init__(self, obj):
        self._obj = obj
        self._class_location = 'mid'


class _FromToMatrix(_LoadHistogramImpl):

    @property
    def index_names(self):
        return set(['from', 'to'])

    def _from_tos(self):
        fr = getattr(self._obj.index.get_level_values('from'), self._class_location).values
        to = getattr(self._obj.index.get_level_values('to'), self._class_location).values
        return fr, to

    def amplitude(self):
        fr, to = self._from_tos()
        return np.abs(fr-to)

    def amplitude_histogram_index(self):
        left = np.zeros(len(self._obj))
        right = np.zeros(len(self._obj))

        fr = self._obj.index.get_level_values('from')
        to = self._obj.index.get_level_values('to')

        hanging = fr.mid > to.mid
        standing = fr.mid <= to.mid

        left[hanging] = fr[hanging].left - to[hanging].right
        right[hanging] = fr[hanging].right - to[hanging].left

        left[standing] = to[standing].left - fr[standing].right
        right[standing] = to[standing].right - fr[standing].left

        left[left < 0.0] = 0.0

        return pd.IntervalIndex.from_arrays(left, right)

    def meanstress(self):
        fr, to = self._from_tos()
        return (fr+to) / 2.


class _RangeMeanMatrix(_LoadHistogramImpl):

    @property
    def index_names(self):
        return set(['range', 'mean'])

    def amplitude(self, location=None):
        return getattr(self._obj.index.get_level_values('range'), location or self._class_location)

    def amplitude_histogram_index(self):
        left = self.amplitude(location='left') / 2.
        right = self.amplitude(location='right') / 2.
        return pd.IntervalIndex.from_arrays(left, right)

    def meanstress(self):
        if 'mean' not in self._obj.index.names:
            return np.zeros_like(self._obj)

        return getattr(self._obj.index.get_level_values('mean'), self._class_location)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Integration tests for the OdbClient

Note that these tests are not part of pyLife's usual CI test pipeline as they
need an Abaqus installation to run.
"""

import os
import pytest
import json

import numpy as np
import pandas as pd

import odbclient

from odbclient.odbclient import OdbServerError

@pytest.fixture
def client():
    return odbclient.OdbClient('tests/beam_3d_hex_quad.odb')


def test_not_existing_odbserver_env():
    with pytest.raises(OSError, match="No odbserver environment found."):
        odbclient.OdbClient('foo.odb', python_env_path='/foo/bar/env')


def test_not_existing_abaqus_path():
    with pytest.raises(FileNotFoundError):
        odbclient.OdbClient('foo.odb', abaqus_bin='/foo/bar/abaqus')


def test_odbclient_instances(client):
    np.testing.assert_array_equal(client.instance_names(), ['PART-1-1'])


def test_odbclient_node_coordinates(client):
    expected = pd.read_csv('tests/node_coordinates.csv', index_col='node_id')
    pd.testing.assert_frame_equal(client.node_coordinates('PART-1-1'), expected)


def test_odbclient_node_ids(client):
    result = client.node_ids('FIX', 'PART-1-1')
    expected = pd.Index([1, 2, 3, 4, 22, 27, 31, 32], dtype='int64', name='node_id')

    pd.testing.assert_index_equal(result, expected)


def test_odbclient_element_ids(client):
    result = client.element_ids('FIX', 'PART-1-1')
    expected = pd.Index([1], dtype='int64', name='element_id')

    pd.testing.assert_index_equal(result, expected)


def test_odbclient_node_coordinates_invalid_instance_name(client):
    with pytest.raises(KeyError, match="Invalid instance name 'nonexistent'."):
        client.node_coordinates('nonexistent')


@pytest.mark.parametrize('instance_name, expected', [
    ('', ' ALL NODES'),
    ('PART-1-1', ['ALL', 'FIX', 'LOAD'])
])
def test_odbclient_nset_names(client, instance_name, expected):
    np.testing.assert_array_equal(client.nset_names(instance_name), expected)


@pytest.mark.timeout(10)
def test_odbclient_nset_names_invalid_instance_name(client):
    with pytest.raises(KeyError, match="Invalid instance name 'nonexistent'."):
        client.nset_names('nonexistent')


@pytest.mark.parametrize('instance_name, expected', [
    ('', ' ALL ELEMENTS'),
    ('PART-1-1', ['ALL', 'FIX'])
])
def test_odbclient_elset_names(client, instance_name, expected):
    np.testing.assert_array_equal(client.elset_names(instance_name), expected)


@pytest.mark.timeout(10)
def test_odbclient_elset_names_invalid_instance_name(client):
    with pytest.raises(KeyError, match="Invalid instance name 'nonexistent'."):
        client.elset_names('nonexistent')


@pytest.mark.skip("to be implemented")
def test_element_connectivity(client):
    expected = pd.read_csv('tests/connectivity.csv', index_col='element_id')
    result = client.element_connectivity('PART-1-1')
    print(result)
    print(expected)
    pd.testing.assert_frame_equal(result, expected)


def test_step_names(client):
    expected = ['Load']
    result = client.step_names()
    np.testing.assert_array_equal(result, expected)


def test_frame_ids(client):
    expected = [0, 1]
    result = client.frame_ids('Load')
    np.testing.assert_array_equal(result, expected)


def test_frame_ids_invalid_step_name(client):
    with pytest.raises(KeyError, match='nonexistent'):
        client.frame_ids('nonexistent')


def test_variable_names(client):
    expected = ['CF', 'COORD', 'E', 'EVOL', 'IVOL', 'RF', 'S', 'U']
    result = client.variable_names('Load', 0)
    np.testing.assert_array_equal(result, expected)


def test_variable_stress_element_nodal(client):
    expected = pd.read_csv('tests/stress_element_nodal.csv', index_col=['node_id', 'element_id'])
    result = client.variable('S', 'PART-1-1', 'Load', 1)

    pd.testing.assert_frame_equal(result, expected)


def test_variable_evol(client):
    result = client.variable('EVOL', 'PART-1-1', 'Load', 1)
    expected = pd.DataFrame({'EVOL': [1000.]}, index=pd.Index([1, 2, 3, 4], name='element_id'))

    pd.testing.assert_frame_equal(result, expected)


def test_variable_invalid_instance_name(client):
    with pytest.raises(KeyError, match="nonexistent"):
        client.variable('S', 'nonexistent', 'Load', 1)


def test_variable_stress_integration_point(client):
    expected = pd.read_csv('tests/stress_integration_point.csv',
                           index_col=['element_id', 'ipoint_id'])
    result = client.variable('S', 'PART-1-1', 'Load', 1, position='INTEGRATION POINTS')
    result.to_csv('tests/stress_integration_point.csv')
    pd.testing.assert_frame_equal(result, expected)

@pytest.fixture
def client_history():
    return odbclient.OdbClient('tests/history_output_test.odb')

def test_history_region_empty(client):
    assert client.history_regions("Load") == ['Assembly ASSEMBLY']

def test_history_region_non_empty(client_history):
    assert client_history.history_regions("Step-1") == ['Assembly ASSEMBLY', 'Element ASSEMBLY.1', 'Node ASSEMBLY.1', 'Node ASSEMBLY.2']

def test_history_outputs(client_history):
    assert client_history.history_outputs("Step-1", 'Element ASSEMBLY.1') == ['CTF1', 'CTF2', 'CTF3', 'CTM1', 'CTM2', 'CTM3', 'CU1', 'CU2', 'CU3', 'CUR1', 'CUR2', 'CUR3']

def test_history_output_values(client_history):
    assert client_history.history_output_values("Step-1", 'Element ASSEMBLY.1', 'CTF1').array[1] == pytest.approx(0.09999854117631912)

def test_history_region_description(client_history):
    assert client_history.history_region_description("Step-1", 'Element ASSEMBLY.1') == "Output at assembly ASSEMBLY instance ASSEMBLY element 1"


def test_history_info(client_history):
    expected = json.loads("""
{"Output at assembly ASSEMBLY instance ASSEMBLY node 1 region RP-1": {"History Outputs": ["RF1", "RF2", "RF3", "RM1", "RM2", "RM3", "U1", "U2", "U3", "UR1", "UR2", "UR3"], "History Region": "Node ASSEMBLY.1", "Steps ": ["Step-1", "Step-2"]}, "Output at assembly ASSEMBLY instance ASSEMBLY element 1": {"History Outputs": ["CTF1", "CTF2", "CTF3", "CTM1", "CTM2", "CTM3", "CU1", "CU2", "CU3", "CUR1", "CUR2", "CUR3"], "History Region": "Element ASSEMBLY.1", "Steps ": ["Step-1", "Step-2"]}, "Output at assembly ASSEMBLY instance ASSEMBLY node 2 region SET-5": {"History Outputs": ["RF1", "RF2", "RF3", "RM1", "RM2", "RM3", "U1", "U2", "U3", "UR1", "UR2", "UR3"], "History Region": "Node ASSEMBLY.2", "Steps ": ["Step-1", "Step-2"]}, "Output at assembly ASSEMBLY": {"History Outputs": ["ALLAE", "ALLCCDW", "ALLCCE", "ALLCCEN", "ALLCCET", "ALLCCSD", "ALLCCSDN", "ALLCCSDT", "ALLCD", "ALLDMD", "ALLDTI", "ALLEE", "ALLFD", "ALLIE", "ALLJD", "ALLKE", "ALLKL", "ALLPD", "ALLQB", "ALLSD", "ALLSE", "ALLVD", "ALLWK", "ETOTAL"], "History Region": "Assembly ASSEMBLY", "Steps ": ["Step-1", "Step-2"]}}
    """)
    assert client_history.history_info() == expected



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pytest
import numpy as np
import pandas as pd
import pylife.stress.rainflow.recorders as RFR


def test_abstract_recorder_new_no_chunks():
    arr = RFR.AbstractRecorder()
    assert len(arr.chunks) == 0


def test_abstract_recorder_one_chunk():
    arr = RFR.AbstractRecorder()
    arr.report_chunk(23)

    np.testing.assert_array_equal(arr.chunks, [23])


def test_abstract_recorder_two_chunks():
    arr = RFR.AbstractRecorder()
    arr.report_chunk(42)
    arr.report_chunk(23)
    np.testing.assert_array_equal(arr.chunks, [42, 23])


def test_abstract_recorder_calc_chunk_local_index():
    arr = RFR.AbstractRecorder()
    arr.report_chunk(42)
    arr.report_chunk(23)

    assert arr.chunk_local_index(0) == (0, 0)
    assert arr.chunk_local_index(17) == (0, 17)
    assert arr.chunk_local_index(42) == (1, 0)
    assert arr.chunk_local_index(47) == (1, 5)


def test_abstract_recorder_calc_chunk_local_index_array():
    arr = RFR.AbstractRecorder()
    arr.report_chunk(42)
    arr.report_chunk(23)

    chunks, index = arr.chunk_local_index([0, 17, 42, 47])
    np.testing.assert_array_equal(chunks, [0, 0, 1, 1])
    np.testing.assert_array_equal(index, [0, 17, 0, 5])


def test_loopvalue_recorder_new_all_empty():
    lvr = RFR.LoopValueRecorder()
    assert len(lvr.values_from) == 0
    assert len(lvr.values_to) == 0


def test_full_rainflow_recorder_new_all_empty():
    fr = RFR.FullRecorder()
    assert len(fr.values_from) == 0
    assert len(fr.values_to) == 0
    assert len(fr.index_from) == 0
    assert len(fr.index_to) == 0


@pytest.mark.parametrize('value_from, value_to, index_from, index_to', [
    (23., 42., 11, 17),
    (46., 84., 22, 34)
])
def test_full_rainflow_recorder_record_value(value_from, value_to, index_from, index_to):
    fr = RFR.FullRecorder()
    fr.record_values([value_from], [value_to])
    fr.record_index([index_from], [index_to])
    np.testing.assert_array_equal(fr.values_from, [value_from])
    np.testing.assert_array_equal(fr.values_to, [value_to])
    np.testing.assert_array_equal(fr.index_from, [index_from])
    np.testing.assert_array_equal(fr.index_to, [index_to])


def test_loop_value_rainflow_recorder_record_two_values():
    vf1, vt1, vf2, vt2 = 23., 42., 46., 84.
    lvr = RFR.LoopValueRecorder()

    lvr.record_values([vf1], [vt1])
    lvr.record_values([vf2], [vt2])

    np.testing.assert_array_equal(lvr.values_from, [vf1, vf2])
    np.testing.assert_array_equal(lvr.values_to, [vt1, vt2])


def test_full_rainflow_recorder_record_two_values():
    vf1, vt1, if1, it1, vf2, vt2, if2, it2 = 23., 42., 11, 17, 46., 84., 22, 34
    fr = RFR.FullRecorder()

    fr.record_values([vf1], [vt1])
    fr.record_index([if1], [it1])

    fr.record_values([vf2], [vt2])
    fr.record_index([if2], [it2])

    np.testing.assert_array_equal(fr.values_from, [vf1, vf2])
    np.testing.assert_array_equal(fr.values_to, [vt1, vt2])
    np.testing.assert_array_equal(fr.index_from, [if1, if2])
    np.testing.assert_array_equal(fr.index_to, [it1, it2])


def test_full_rainflow_recorder_empty_collective_default():
    fr = RFR.FullRecorder()
    expected = pd.DataFrame({
        'from': [],
        'to': [],
        'index_from': pd.Series([], dtype=np.uintp),
        'index_to': pd.Series([], dtype=np.uintp)
    })
    pd.testing.assert_frame_equal(fr.collective, expected)


def test_full_rainflow_recorder_two_non_zero_collective():
    vf1, vt1, if1, it1, vf2, vt2, if2, it2 = 23., 42., 11, 17, 46., 84., 22, 34
    fr = RFR.FullRecorder()
    fr.record_values([vf1], [vt1])
    fr.record_index([if1], [it1])
    fr.record_values([vf2], [vt2])
    fr.record_index([if2], [it2])

    expected = pd.DataFrame({
        'from': [vf1, vf2],
        'to': [vt1, vt2],
        'index_from': pd.Series([if1, if2], dtype=np.uintp),
        'index_to': pd.Series([it1, it2], dtype=np.uintp)
    })

    print(expected)
    print(expected.dtypes)
    pd.testing.assert_frame_equal(fr.collective, expected)


def test_full_rainflow_recorder_empty_histogram_default():
    fr = RFR.FullRecorder()
    histogram, _, _ = fr.histogram_numpy()
    np.testing.assert_array_equal(histogram, np.zeros((10, 10)))


def test_full_rainflow_recorder_empty_histogram_5_bins():
    fr = RFR.FullRecorder()
    histogram, _, _ = fr.histogram_numpy(bins=5)
    np.testing.assert_array_equal(histogram, np.zeros((5, 5)))


@pytest.mark.parametrize('value_from, value_to, index_from, index_to', [
    (23., 42., 11, 17),
    (46., 84., 22, 34)
])
def test_full_rainflow_recorder_one_non_zero(value_from, value_to, index_from, index_to):
    fr = RFR.FullRecorder()
    fr.record_values([value_from], [value_to])
    fr.record_index([index_from], [index_to])

    expected_from = np.linspace(value_from - 0.5, value_from + 0.5, 11)
    expected_to = np.linspace(value_to - 0.5, value_to + 0.5, 11)
    expected_histogram = np.zeros((10, 10))
    expected_histogram[5, 5] = 1.

    histogram, vfrom, vto = fr.histogram_numpy()
    np.testing.assert_array_equal(expected_from, vfrom)
    np.testing.assert_array_equal(expected_to, vto)
    np.testing.assert_array_equal(expected_histogram, histogram)


def test_full_rainflow_recorder_two_non_zero():
    vf1, vt1, if1, it1, vf2, vt2, if2, it2 = 23., 42., 11, 17, 46., 84., 22, 34
    fr = RFR.FullRecorder()
    fr.record_values([vf1], [vt1])
    fr.record_index([if1], [it1])
    fr.record_values([vf2], [vt2])
    fr.record_index([if2], [it2])
    fr.record_values([vf2], [vt2])
    fr.record_index([if2], [it2])

    expected_from = np.linspace(vf1, vf2, 11)
    expected_to = np.linspace(vt1, vt2, 11)
    expected_histogram = np.zeros((10, 10))
    expected_histogram[0, 0] = 1.
    expected_histogram[-1, -1] = 2.

    histogram, vfrom, vto = fr.histogram_numpy()
    np.testing.assert_array_equal(expected_from, vfrom)
    np.testing.assert_array_equal(expected_to, vto)
    np.testing.assert_array_equal(expected_histogram, histogram)


def test_loopvalue_rainflow_recorder_empty_collective_default():
    lvr = RFR.LoopValueRecorder()
    expected = pd.DataFrame({'from': [], 'to': []})
    pd.testing.assert_frame_equal(lvr.collective, expected)


def test_loop_value_rainflow_recorder_record_two_values_collective():
    vf1, vt1, vf2, vt2 = 23., 42., 46., 84.
    lvr = RFR.LoopValueRecorder()

    lvr.record_values([vf1], [vt1])
    lvr.record_values([vf2], [vt2])

    expected = pd.DataFrame({
        'from': [vf1, vf2],
        'to': [vt1, vt2]
    })

    pd.testing.assert_frame_equal(lvr.collective, expected)


def test_lopvalue_rainflow_recorder_empty_histogram_default():
    fr = RFR.LoopValueRecorder()
    histogram = fr.histogram()
    assert histogram.index.names[0] == 'from'
    assert histogram.index.names[1] == 'to'
    np.testing.assert_array_equal(histogram.to_numpy(), np.zeros(100))


def test_loopvalue_rainflow_recorder_empty_histogram_5_bins():
    fr = RFR.LoopValueRecorder()
    histogram = fr.histogram(bins=5).to_numpy()
    np.testing.assert_array_equal(histogram, np.zeros(25))


@pytest.mark.parametrize('value_from, value_to', [
    (23., 42.),
    (46., 84.)
])
def test_loopvalue_rainflow_recorder_histogram_one_non_zero(value_from, value_to):
    fr = RFR.LoopValueRecorder()
    fr.record_values([value_from], [value_to])

    expected_from = pd.IntervalIndex.from_breaks(np.linspace(value_from - 0.5, value_from + 0.5, 11))
    expected_to = pd.IntervalIndex.from_breaks(np.linspace(value_to - 0.5, value_to + 0.5, 11))
    expected_index = pd.MultiIndex.from_product([expected_from, expected_to], names=['from', 'to'])

    expected_histogram = np.zeros((10, 10))
    expected_histogram[5, 5] = 1.

    histogram = fr.histogram()
    pd.testing.assert_series_equal(histogram, pd.Series(expected_histogram.flatten(), index=expected_index))



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import sys
import os

import pickle

import numpy as np

from .interface import OdbInterface

class OdbServer:

    def __init__(self, odbfile):
        self._odb = OdbInterface(odbfile)
        self.command_dict = {
            "get_instances": self.instances,
            "get_steps": self.steps,
            "get_frames": self.frames,
            "get_nodes": self.nodes,
            "get_connectivity": self.connectivity,
            "get_node_sets": self.node_sets,
            "get_element_sets": self.element_sets,
            "get_node_set": self.node_set,
            "get_element_set": self.element_set,
            "get_variable_names": self.variable_names,
            "get_variable": self.variable,
            "get_history_regions": self.history_regions,
            "get_history_outputs": self.history_outputs,
            "get_history_output_values": self.history_output_values,
            "get_history_region_description": self.history_region_description,
            "get_history_info": self.history_info
        }

    def instances(self, _args):
        _send_response(self._odb.instance_names())

    def steps(self, _args):
        _send_response(self._odb.step_names())

    def frames(self, step_name):
        _send_response(self._odb.frame_names(str(step_name)))

    def nodes(self, args):
        instance_name, node_set_name = args
        try:
            nodes = self._odb.nodes(str(instance_name), str(node_set_name))
        except Exception as e:
            _send_response(e)
        else:
            _send_response(nodes)

    def connectivity(self, args):
        instance_name, element_set_name = args
        try:
            conn = self._odb.connectivity(str(instance_name), str(element_set_name))
        except Exception as e:
            _send_response(e)
        else:
            _send_response(conn)

    def node_sets(self, instance_name):
        _send_response(self._odb.node_sets(str(instance_name)))

    def element_sets(self, instance_name):
        _send_response(self._odb.element_sets(str(instance_name)))

    def node_set(self, args):
        instance_name, node_set_name = args
        _send_response(self._odb.node_set(str(instance_name), str(node_set_name)))

    def element_set(self, args):
        instance_name, element_set_name = args
        _send_response(self._odb.element_set(str(instance_name), str(element_set_name)))

    def variable_names(self, args):
        step, frame = args
        _send_response(self._odb.variable_names(step, int(frame)))

    def variable(self, args):
        instance_name, step, frame, var_name, nset, elset, elem_pos = args
        instance_name = str(instance_name)
        var_name = str(var_name)
        nset = str(nset)
        elset = str(elset)
        try:
            variable = self._odb.variable(instance_name, step, frame, var_name, nset, elset, elem_pos)
        except Exception as e:
            _send_response(e)
        else:
            _send_response(variable)

    def history_regions(self, step_name):
        _send_response(self._odb.history_regions(str(step_name)))

    def history_outputs(self, args):
        step_name, historyregion_name = args
        step_name = str(step_name)
        historyregion_name = str(historyregion_name)
        _send_response(self._odb.history_outputs(step_name, historyregion_name))

    def history_output_values(self, args):
        step_name, historyregion_name, historyoutput_name = args
        step_name = str(step_name)
        historyregion_name = str(historyregion_name)
        historyoutput_name = str(historyoutput_name)
        _send_response(self._odb.history_output_values(step_name, historyregion_name, historyoutput_name))

    def history_region_description(self, args):
        step_name, historyregion_name = args
        step_name = str(step_name)
        historyregion_name = str(historyregion_name)
        _send_response(self._odb.history_region_description(step_name, historyregion_name))

    def history_info(self, args):
        _send_response(self._odb.history_info())

def _send_response(pickle_data, numpy_arrays=None):
    numpy_arrays = numpy_arrays or []
    s = pickle.dumps((len(numpy_arrays), pickle_data))
    sys.stdout.write(s + '\n')
    sys.stdout.flush()
    for nparr in numpy_arrays:
        np.lib.format.write_array(sys.stdout, nparr)


def main():

    odbfile = sys.argv[1]

    try:
        server = OdbServer(odbfile)
    except Exception as e:
        print >> sys.stderr, e.message
        sys.exit(1)

    sys.stdout.write('ready')
    sys.stdout.flush()

    command = ''
    while True:
        command, parameters = pickle.load(sys.stdin)

        if command == 'QUIT':
            break

        func = server.command_dict.get(command)
        if func is not None:
            func(parameters)
            sys.stdout.flush()


if __name__ == "__main__":
    main()



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__


import pandas as pd

from .broadcaster import Broadcaster
from .data_validator import DataValidator


class PylifeSignal(Broadcaster):
    """Base class for signal accessor classes.

    Notes
    -----
    Derived classes need to implement the method `_validate(self, obj)`
    that gets `pandas_obj` as `obj` parameter. This `validate()` method
    must raise an Exception (e.g. AttributeError or ValueError) in case
    `obj` is not a valid DataFrame for the kind of signal.

    For these validation :func:`fail_if_key_missing()` and
    :func:`get_missing_keys()` might be helpful.

    For a derived class you can register methods without modifying the
    class' code itself. This can be useful if you want to make signal
    accessor classes extendable.

    See also
    --------
    :func:`fail_if_key_missing()`
    :func:`get_missing_keys()`
    :func:`register_method()`
    """

    _method_dict = {}

    def __init__(self, pandas_obj):
        """Instantiate a :class:`signal.PyLifeSignal`.

        Parameters
        ----------
        pandas_obj : pandas.DataFrame or pandas.Series

        """
        self._obj = pandas_obj
        self._validate()

    @classmethod
    def from_parameters(cls, **kwargs):
        """Make a signal instance from a parameter set.

        This is a convenience function to instantiate a signal from individual
        parameters rather than pandas objects.

        A signal class like

        .. code-block:: python

            @pd.api.extensions.register_dataframe_accessor('foo_signal')
            class FooSignal(PylifeSignal):
                pass

        The following two blocks are equivalent:

        .. code-block:: python

            pd.Series({'foo': 1.0, 'bar': 2.0}).foo_signal

        .. code-block:: python

            FooSignal.from_parameters(foo=1.0, bar=1.0)

        """
        # TODO: better error handling
        if len(kwargs) > 1 and hasattr(next(iter(kwargs.values())), '__iter__'):
            obj = pd.DataFrame(kwargs)
        else:
            obj = pd.Series(kwargs)

        return cls(obj)

    def keys(self):
        """Get a list of missing keys that are needed for a signal object.

        Returns
        -------
        keys : pd.Index
            a pandas index of keys

        Raises
        ------
        AttributeError
            if `self._obj` is neither a `pandas.DataFrame` nor a `pandas.Series`

        Notes
        -----
        If `self._obj` is a `pandas.DataFrame`, the `self._obj.columns` are returned.

        If `self._obj` is a `pandas.Series`, the `self._obj.index` are returned.
        """
        if isinstance(self._obj, pd.Series):
            return self._obj.index
        elif isinstance(self._obj, pd.DataFrame):
            return self._obj.columns
        raise AttributeError("An accessor object needs to be either a pandas.Series or a pandas.DataFrame")

    def get_missing_keys(self, keys_to_check):
        """Get a list of missing keys that are needed for a self._obj object.

        Parameters
        ----------
        keys_to_check : list
            A list of keys that need to be available in `self._obj`

        Returns
        -------
        missing_keys : list
            a list of missing keys

        Raises
        ------
        AttributeError
            if `self._obj` is neither a `pandas.DataFrame` nor a `pandas.Series`

        Notes
        -----
        If `self._obj` is a `pandas.DataFrame`, all keys of
        `keys_to_check` not found in the `self._obj.columns` are
        returned.

        If `self._obj` is a `pandas.Series`, all keys of
        `keys_to_check` not found in the `self._obj.index` are
        returned.
        """
        return DataValidator().get_missing_keys(self._obj, keys_to_check)

    def fail_if_key_missing(self, keys_to_check, msg=None):
        """Raise an exception if any key is missing in a self._obj object.

        Parameters
        ----------
        self._obj : pandas.DataFrame or pandas.Series
            The object to be checked

        keys_to_check : list
            A list of keys that need to be available in `self._obj`

        Raises
        ------
        AttributeError
            if `self._obj` is neither a `pandas.DataFrame` nor a `pandas.Series`
        AttributeError
            if any of the keys is not found in the self._obj's keys.

        Notes
        -----
        If `self._obj` is a `pandas.DataFrame`, all keys of
        `keys_to_check` meed to be found in the `self._obj.columns`.

        If `self._obj` is a `pandas.Series`, all keys of
        `keys_to_check` meed to be found in the `self._obj.index`.

        See also
        --------
        :func:`get_missing_keys`
        :class:`stresssignal.StressTensorVoigt`
        """
        DataValidator().fail_if_key_missing(self._obj, keys_to_check)

    class _MethodCaller:
        def __init__(self, method, obj):
            self._method = method
            self._obj = obj

        def __call__(self, *args, **kwargs):
            return self._method(self._obj, *args, **kwargs)

    def __getattr__(self, itemname):
        method = self._method_dict.get(itemname)

        if method is None:
            return super().__getattribute__(itemname)

        return self._MethodCaller(method, self._obj)

    @classmethod
    def _register_method(cls, method_name):
        def method_decorator(method):
            if method_name in cls._method_dict.keys():
                raise ValueError("Method '%s' already registered in %s" % (method_name, cls.__name__))
            if hasattr(cls, method_name):
                raise ValueError("%s already has an attribute '%s'" % (cls.__name__, method_name))
            cls._method_dict[method_name] = method
        return method_decorator

    def to_pandas(self):
        """Expose the pandas object of the signal.

        Returns
        -------
        pandas_object : pd.DataFrame or pd.Series
            The pandas object representing the signal


        Notes
        -----

        The default implementation just returns the object given when
        instantiating the signal class. Derived classes may return a modified
        object or augmented, if they store some extra information.

        By default the object is **not** copied. So make a copy yourself, if
        you intent to modify it.
        """
        return self._obj


def register_method(cls, method_name):
    """Registers a method to a class derived from :class:`PyifeSignal`

    Parameters
    ----------
    cls : class
        The class the method is registered to.
    method_name : str
        The name of the method

    Raises
    ------
    ValueError
        if `method_name` is already registered for the class
    ValueError
        if `method_name` the class has already an attribute `method_name`

    Notes
    -----
    The function is meant to be used as a decorator for a function
    that is to be installed as a method for a class. The class is
    assumed to contain a pandas object in `self._obj`.

    Examples
    --------
    .. code-block:: python

        import pandas as pd
        import pylife as pl


        @pd.api.extensions.register_dataframe_accessor('foo')
        class Foo(pl.signal.PyifeSignal):
            def __init__(self, obj):
                # self._validate(obj) could come here
                self._obj = obj


        @pl.signal.register_method(Foo, 'bar')
        def bar(df):
            return pd.DataFrame({'baz': df['foo'] + df['bar']})

    >>> df = pd.DataFrame({'foo': [1.0, 2.0], 'bar': [-1.0, -2.0]})
    >>> df.foo.bar()
       baz
    0  0.0
    1  0.0
    """
    return cls._register_method(method_name)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np
import pandas as pd

import pylife.strength.meanstress as MST


@pytest.fixture
def ms_sens():
    return pd.Series({'M': 0.5, 'M2': 0.5/3.})


def test_meanstress_collective_empty_fail():
    df = pd.DataFrame({'foo': [], 'bar': []})
    with pytest.raises(AttributeError, match="Load collective"):
        df.meanstress_transform


def test_meanstress_collective_empty_fkm_goodman(ms_sens):
    df = pd.DataFrame(columns=['from', 'to'])

    res = df.meanstress_transform.fkm_goodman(ms_sens, -1.).to_pandas()
    assert 'from' in res.columns
    assert 'to' in res.columns
    assert len(res.columns) == 2
    assert res.shape[0] == 0


def test_meanstress_collective_fkm_goodman_single_ms_sens(ms_sens):
    df = pd.DataFrame({
        'from': [-6., -4.,  -5./2., -1., -0.4, 0., 7./12.],
        'to': [-2., 0., 0.5, 1., 1.2, 4./3., 21./12.]
    }, index=[3, 4, 5, 6, 7, 8, 9])

    res = df.meanstress_transform.fkm_goodman(ms_sens, -1.)

    expected_amplitude = pd.Series(np.ones(7), name='amplitude', index=df.index)
    pd.testing.assert_series_equal(res.amplitude, expected_amplitude)

    expected_meanstress = pd.Series(np.zeros(7), name='meanstress', index=df.index)
    pd.testing.assert_series_equal(res.meanstress, expected_meanstress)


def test_meanstress_collective_fkm_goodman_multiple_ms_sens():
    df = pd.DataFrame({
        'from': [-6., -4.,  -5./2., -1., -0.4, 0., 7./12.],
        'to': [-2., 0., 0.5, 1., 1.2, 4./3., 21./12.]
    }, index=pd.Index([3, 4, 5, 6, 7, 8, 9], name='element_id'))

    ms_sens = pd.DataFrame({
        'M': [0.5, 0.4],
        'M2': [0.5/3., 0.4/3.]
    })

    expected_index = pd.MultiIndex.from_tuples([
        (0, 3), (0, 4), (0, 5), (0, 6), (0, 7), (0, 8), (0, 9),
        (1, 3), (1, 4), (1, 5), (1, 6), (1, 7), (1, 8), (1, 9)
    ], names=[None, 'element_id'])

    res = df.meanstress_transform.fkm_goodman(ms_sens, -1.)

    expected_amplitude = pd.Series(
        [1., 1., 1., 1., 1., 1., 1., 1.2, 1.2, 1.1, 1.0, 0.96, 0.93, 0.91],
        index=expected_index,
        name='amplitude'
    )
    pd.testing.assert_series_equal(res.amplitude, expected_amplitude, rtol=1e-2)

    expected_meanstress = pd.Series(
        np.zeros(14),
        index=expected_index,
        name='meanstress'
    )
    pd.testing.assert_series_equal(res.meanstress, expected_meanstress, rtol=1e-2)



import re
import pandas as pd
import pytest

import pylife.vmap as vmap
import h5py
import os
from . import reference_data as RD
import pylife.vmap.vmap_structures as structures


class TestExport:
    @pytest.fixture(scope='function', autouse=True)
    def prepare_data(self, tmp_path_factory):
        print("prepare data")
        tmpdir = tmp_path_factory.mktemp('vmap-export').as_posix()

        self._export = vmap.VMAPExport(os.path.join(tmpdir, 'test.vmap'))
        self._import_expected = vmap.VMAPImport('tests/vmap/testfiles/beam_2d_squ_lin.vmap')
        self._mesh = (self._import_expected.make_mesh('1', 'STATE-2')
                      .join_coordinates()
                      .join_variable('DISPLACEMENT')
                      .to_frame())
        self._export.add_geometry('1', self._mesh)

    def test_fundamental_groups(self):
        with vmap.VMAPImport(self._export.file_name) as import_actual:
            vmap_group = import_actual.try_get_vmap_object('VMAP')
            assert vmap_group is not None
            assert len(vmap_group) == 4

            geometry_group = import_actual.try_get_vmap_object('VMAP/GEOMETRY')
            assert geometry_group is not None
            assert len(geometry_group) == 1

            material_group = import_actual.try_get_vmap_object('VMAP/MATERIAL')
            assert material_group is not None
            assert len(material_group) == 0

            system_group = import_actual.try_get_vmap_object('VMAP/SYSTEM')
            assert system_group is not None
            assert len(system_group) == 5
            self.assert_dataset_correct(system_group['ELEMENTTYPES'], self._export._element_types)
            self.assert_dataset_correct(system_group['METADATA'], self._export._metadata, False)
            self.assert_dataset_correct(system_group['SECTION'], self._export._sections)
            self.assert_dataset_correct(system_group['UNITSYSTEM'], self._export._unit_system)
            self.assert_dataset_correct(system_group['COORDINATESYSTEM'], self._export._coordinate_systems)

            variables_group = import_actual.try_get_vmap_object('VMAP/VARIABLES')
            assert variables_group is not None
            assert len(variables_group) == 0

    def test_set_geometry_name(self):
        geometry_path = 'VMAP/GEOMETRY/1'
        name = 'PART-1-1'
        self._export.set_group_attribute(geometry_path, 'MYNAME', name)
        with h5py.File(self._export.file_name, 'r') as file:
            assert file[geometry_path].attrs['MYNAME'] == name

    def test_set_invalid_geometry_name(self):
        geometry_path = 'INVALID'
        name = 'PART-1-1'
        with pytest.raises(KeyError, match='VMAP object INVALID does not exist.'):
            self._export.set_group_attribute(geometry_path, 'MYNAME', name)

    def test_add_dataset(self):
        self._export.add_integration_types(RD.integration_type_content)
        dataset_name = 'INTEGRATIONTYPES'
        with vmap.VMAPImport(self._export.file_name) as import_actual:
            dataset = import_actual.try_get_vmap_object('VMAP/SYSTEM/%s' % dataset_name)
            assert dataset is not None
            self.assert_dataset_correct(dataset, RD.integration_type_content)

    def test_add_dataset_already_exists(self):
        self._export.add_integration_types(RD.integration_type_content)
        with pytest.raises(KeyError):
            self._export.add_integration_types(RD.integration_type_content)

    def test_geometry(self):
        geometry_full_path = "VMAP/GEOMETRY/1"
        with vmap.VMAPImport(self._export.file_name) as import_actual:
            geometry_actual = import_actual.try_get_vmap_object(geometry_full_path)
            assert geometry_actual is not None
            geometry_expected = self._import_expected.try_get_vmap_object(geometry_full_path)
            assert geometry_expected is not None

            elements_actual = import_actual.try_get_vmap_object("%s/ELEMENTS" % geometry_full_path)
            assert elements_actual is not None
            elements_expected = self._import_expected.try_get_vmap_object("%s/ELEMENTS" % geometry_full_path)
            assert elements_expected is not None

            self.assert_group_attrs_equal(elements_expected, elements_actual)

            points_actual = import_actual.try_get_vmap_object("%s/POINTS" % geometry_full_path)
            assert points_actual is not None
            points_expected = self._import_expected.try_get_vmap_object("%s/POINTS" % geometry_full_path)
            assert points_expected is not None

            self.assert_group_attrs_equal(points_expected, points_actual)

    def test_add_geometry(self):
        geometry_name = '2'
        self._export.add_geometry(geometry_name, self._mesh)
        mesh_expected = self._mesh[['x', 'y', 'z']]
        with vmap.VMAPImport(self._export.file_name) as import_actual:
            mesh_actual = (import_actual.make_mesh(geometry_name)
                           .join_coordinates()
                           .to_frame())
            pd.testing.assert_frame_equal(mesh_expected.sort_index(), mesh_actual.sort_index())

    def test_add_geometry_invalid(self):
        geometry_name = '2'
        with pytest.raises(vmap.VMAPExportError):
            self._export.add_geometry(geometry_name, 5)

        with vmap.VMAPImport(self._export.file_name) as import_actual:
            geometry = import_actual.try_get_vmap_object('VMAP/GEOMETRY/%s' % geometry_name)
            assert geometry is None

    def test_add_geometry_already_exists(self):
        geometry_name = '1'
        with pytest.raises(KeyError):
            self._export.add_geometry(geometry_name, self._mesh)

    def test_add_node_set(self):
        geometry_name = '1'
        geometry_set_name = '000000'
        geometry_set_full_path = 'VMAP/GEOMETRY/%s/GEOMETRYSETS/%s' % (geometry_name, geometry_set_name)
        geometry_set_expected = self._import_expected.try_get_geometry_set(geometry_name, geometry_set_name)
        assert geometry_set_expected is not None
        self._export.add_node_set(geometry_name, geometry_set_expected, self._mesh, 'ALL')
        with vmap.VMAPImport(self._export.file_name) as import_actual:
            geometry_set_group_expected = self._import_expected.try_get_vmap_object(geometry_set_full_path)
            assert geometry_set_group_expected is not None
            geometry_set_group_actual = import_actual.try_get_vmap_object(geometry_set_full_path)
            assert geometry_set_group_actual is not None
            self.assert_group_attrs_equal(geometry_set_group_expected, geometry_set_group_actual)

            geometry_set_actual = import_actual.try_get_geometry_set(geometry_name, geometry_set_name)
            assert geometry_set_actual is not None
            assert geometry_set_expected.shape == geometry_set_actual.shape
            assert geometry_set_expected.size == geometry_set_actual.size
            assert geometry_set_expected.dtype == geometry_set_actual.dtype
            assert (geometry_set_expected == geometry_set_actual).all()

    def test_add_element_set(self):
        geometry_name = '1'
        geometry_set_name_expected = '000003'
        geometry_set_name_actual = '000000'
        geometry_set_full_path_expected = 'VMAP/GEOMETRY/%s/GEOMETRYSETS/%s' % (geometry_name,
                                                                                geometry_set_name_expected)
        geometry_set_full_path_actual = 'VMAP/GEOMETRY/%s/GEOMETRYSETS/%s' % (geometry_name,
                                                                              geometry_set_name_actual)
        geometry_set_expected = self._import_expected.try_get_geometry_set(geometry_name, geometry_set_name_expected)
        assert geometry_set_expected is not None
        self._export.add_element_set(geometry_name, geometry_set_expected, self._mesh, 'ALL')
        with vmap.VMAPImport(self._export.file_name) as import_actual:
            geometry_set_group_expected = self._import_expected.try_get_vmap_object(geometry_set_full_path_expected)
            assert geometry_set_group_expected is not None
            geometry_set_group_actual = import_actual.try_get_vmap_object(geometry_set_full_path_actual)
            assert geometry_set_group_actual is not None
            self.assert_group_attrs_equal(geometry_set_group_expected, geometry_set_group_actual, 'MYIDENTIFIER')

            geometry_set_actual = import_actual.try_get_geometry_set(geometry_name, geometry_set_name_actual)
            assert geometry_set_actual is not None
            assert geometry_set_expected.shape == geometry_set_actual.shape
            assert geometry_set_expected.size == geometry_set_actual.size
            assert geometry_set_expected.dtype == geometry_set_actual.dtype
            assert (geometry_set_expected == geometry_set_actual).all()

    def test_add_node_set_invalid(self):
        geometry_name = '1'
        invalid_node_set = pd.Index([17, 18, 19])
        with pytest.raises(KeyError, match='Provided index set is not a subset of the node indices.'):
            self._export.add_node_set(geometry_name, invalid_node_set, self._mesh, 'ALL')

    def test_add_element_set_invalid(self):
        geometry_name = '1'
        invalid_elemment_set = pd.Index([17, 18, 19])
        with pytest.raises(KeyError, match='Provided index set is not a subset of the element indices.'):
            self._export.add_element_set(geometry_name, invalid_elemment_set, self._mesh, 'ALL')

    def test_add_element_set_invalid_name(self):
        geometry_name = '1'
        elemment_set = pd.Index([1, 2, 3])
        invalid_name = 123
        with pytest.raises(TypeError, match=re.escape('Invalid set name (must be a string).')):
            self._export.add_element_set(geometry_name, elemment_set, self._mesh, invalid_name)

    def test_add_element_set_invalid_geometry_name(self):
        geometry_name = 'foo'
        elemment_set = pd.Index([1, 2, 3])
        with pytest.raises(KeyError, match=re.escape('No geometry with the name foo')):
            self._export.add_element_set(geometry_name, elemment_set, self._mesh, 'ALL')

    def test_add_variable(self):
        state_name = 'STATE-2'
        state_full_path = 'VMAP/VARIABLES/%s' % state_name
        geometry_name = '1'
        variable_name = 'DISPLACEMENT'
        variable_full_path = '%s/%s/%s' % (state_full_path, geometry_name, variable_name)
        self._export.add_variable(state_name, geometry_name, variable_name, self._mesh)
        with vmap.VMAPImport(self._export.file_name) as import_actual:
            variable_group_expected = self._import_expected.try_get_vmap_object(variable_full_path)
            assert variable_group_expected is not None
            variable_group_actual = import_actual.try_get_vmap_object(variable_full_path)
            assert variable_group_expected is not None
            self.assert_group_attrs_equal(variable_group_expected, variable_group_actual,
                                          'MYIDENTIFIER', 'MYTIMEVALUE', 'MYVARIABLEDESCRIPTION')
            mesh_actual = (import_actual.make_mesh(geometry_name, state_name)
                           .join_variable(variable_name)
                           .to_frame())
        mesh_expected = self._mesh[self._export.variable_column_names(variable_name)]
        pd.testing.assert_frame_equal(mesh_expected.sort_index(), mesh_actual.sort_index())

    def test_add_variable_group_invalid(self):
        state_name = 'STATE-2'
        geometry_name = '2'
        variable_name = 'DISPLACEMENT'
        with pytest.raises(KeyError):
            self._export.add_variable(state_name, geometry_name, variable_name, self._mesh)

    def test_add_variable_name_invalid(self):
        state_name = 'STATE-2'
        geometry_name = '1'
        variable_name = 'DISPLACEMENT2'
        with pytest.raises(KeyError):
            self._export.add_variable(state_name, geometry_name, variable_name, self._mesh)

    def test_add_variable_unknown_location(self):
        state_name = 'STATE-2'
        geometry_name = '1'
        variable_name = 'FORCE_REACTION'
        mesh = (self._import_expected.make_mesh(geometry_name, state_name)
                .join_variable(variable_name, column_names=['RF1', 'RF2', 'RF3'])
                .to_frame())
        with pytest.raises(vmap.APIUseError,
                           match=re.escape(
                               "Need location for unknown variable RF1. "
                               "Please provide one using 'location' parameter.")):
            self._export.add_variable(state_name, geometry_name, 'RF1', mesh, column_names=['RF1'])

    def test_add_variable_unknown_column_name(self):
        state_name = 'STATE-2'
        geometry_name = '1'
        variable_name = 'FORCE_REACTION'
        mesh_expected = (self._import_expected.make_mesh(geometry_name, state_name)
                         .join_variable(variable_name, column_names=['RF1', 'RF2', 'RF3'])
                         .to_frame())
        self._export.add_variable(state_name, geometry_name, 'RF1', mesh_expected, column_names=['RF1'],
                                  location=structures.VariableLocations.NODE)
        with vmap.VMAPImport(self._export.file_name) as import_actual:
            mesh_actual = (import_actual.make_mesh(geometry_name, state_name)
                           .join_variable('RF1', column_names=['RF1'])
                           .to_frame())
            pd.testing.assert_series_equal(mesh_actual['RF1'], mesh_expected['RF1'])

    def test_add_variable_not_present_in_mesh(self):
        state_name = 'STATE-2'
        geometry_name = '1'
        variable_name = 'FORCE_REACTION'
        with pytest.raises(vmap.VMAPExportError):
            self._export.add_variable(state_name, geometry_name, variable_name, self._mesh,
                                      column_names=['RF1'], location=structures.VariableLocations.NODE)
        with vmap.VMAPImport(self._export.file_name) as import_actual:
            variable = import_actual.try_get_vmap_object(
                'VMAP/VARIABLES/%s/%s/%s' % (state_name, geometry_name, variable_name))
            assert variable is None

    def test_add_variable_invalid_location(self):
        state_name = 'STATE-2'
        geometry_name = '1'
        variable_name = 'DISPLACEMENT'
        with pytest.raises(vmap.APIUseError,
                           match=re.escape(
                               "location parameter needs to be of type VariableLocations.")):
            self._export.add_variable(state_name, geometry_name, variable_name, self._mesh,
                                      column_names=['dx', 'dy', 'dz'], location=4)
        with vmap.VMAPImport(self._export.file_name) as import_actual:
            variable = import_actual.try_get_vmap_object(
                'VMAP/VARIABLES/%s/%s/%s' % (state_name, geometry_name, variable_name))
            assert variable is None

    def test_variable_location_displacement(self):
        assert self._export.variable_location('DISPLACEMENT') == structures.VariableLocations.NODE

    def test_variable_location_stress_cauchy(self):
        assert self._export.variable_location('STRESS_CAUCHY') == structures.VariableLocations.ELEMENT_NODAL

    def test_add_variable_already_exists(self):
        state_name = 'STATE-2'
        geometry_name = '1'
        variable_name = 'DISPLACEMENT'
        self._export.add_variable(state_name, geometry_name, variable_name, self._mesh)
        with pytest.raises(KeyError):
            self._export.add_variable(state_name, geometry_name, variable_name, self._mesh)

    """
    def test_all(self):
        self.test_add_dataset()
        self.test_add_geometry()
        self.test_add_node_set()
        self.test_add_variable()
    """

    def assert_dataset_correct(self, dataset, expected_values, is_compound=True):
        assert len(dataset) == len(expected_values)
        for e_key, a_0 in zip(expected_values, dataset):
            e_0 = expected_values[e_key]
            if is_compound:
                a_0 = a_0[0]
            assert len(e_0) == len(a_0)
            for e_1, a_1 in zip(e_0, a_0):
                if isinstance(e_1, list):
                    for e_2, a_2 in zip(e_1, a_1):
                        assert e_2 == a_2
                    continue
                e_1 = self.make_bytearray_if_str(e_1)
                a_1 = self.make_bytearray_if_str(a_1)
                assert e_1 == a_1

    def assert_group_attrs_equal(self, group_expected, group_actual, *args):
        attributes_expected = list(group_expected.attrs.items())
        for attr_expected in attributes_expected:
            assert attr_expected[0] in group_actual.attrs
            if attr_expected[0] not in args:
                testval = self.make_bytearray_if_str(group_actual.attrs[attr_expected[0]])
                assert testval == attr_expected[1]

    def make_bytearray_if_str(self, value):
        if isinstance(value, str):
            return bytearray(value, 'utf-8')
        return value


@pytest.mark.parametrize('filename', [
    'beam_2d_tri_lin.vmap',
    'beam_2d_tri_quad.vmap',
    'beam_2d_squ_lin.vmap',
    'beam_2d_squ_quad.vmap',
    'beam_3d_tet_lin.vmap',
    'beam_3d_tet_quad.vmap',
    'beam_3d_wedge_lin.vmap',
    'beam_3d_wedge_quad.vmap',
    'beam_3d_hex_lin.vmap',
    'beam_3d_hex_quad.vmap',
])
def test_export_import_round_robin(tmpdir, filename):
    filename = os.path.join('tests/vmap/testfiles/', filename)
    import_expected = vmap.VMAPImport(filename)
    mesh = (import_expected.make_mesh('1', 'STATE-2')
            .join_coordinates()
            .join_variable('STRESS_CAUCHY')
            .join_variable('DISPLACEMENT')
            .join_variable('E')
            .to_frame())

    export_filename = os.path.join(tmpdir, 'export.vmap')
    (vmap.VMAPExport(export_filename)
     .add_geometry('1', mesh)
     .add_variable('STATE-2', '1', 'STRESS_CAUCHY', mesh)
     .add_variable('STATE-2', '1', 'DISPLACEMENT', mesh)
     .add_variable('STATE-2', '1', 'E', mesh))

    reimport = vmap.VMAPImport(export_filename)
    reimported_mesh = (reimport.make_mesh('1', 'STATE-2')
                       .join_coordinates()
                       .join_variable('STRESS_CAUCHY')
                       .join_variable('DISPLACEMENT')
                       .join_variable('E')
                       .to_frame())

    pd.testing.assert_frame_equal(mesh, reimported_mesh)


def test_os_error_on_open_file():
    with pytest.raises(FileNotFoundError):
        self._export = vmap.VMAPExport(os.path.join("/some/most/probably/not/existing/path", "test.vmap"))



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 16 10:24:26 2021

@author: SIS3SI
"""
import re
import pytest
import numpy as np

from  pylife.materiallaws.true_stress_strain import *

parametrization_data_strain = np.array([
    [0.0, 0.0],
    [1.0, 0.6931471805599453],
    [2.0, 1.0986122886681098]
])

@pytest.mark.parametrize('tech_strain, expected', map(tuple, parametrization_data_strain))
def test_true_strain_scalar(tech_strain, expected):
    np.testing.assert_approx_equal(true_strain(tech_strain), expected, significant=5)
    
    
parametrization_data_stress = np.array([
    [0.0, 0.0, 0.0],
    [1.0, 1.0, 2.0],
    [2.0, 2.0, 6.0]
])    

@pytest.mark.parametrize('tech_stress, tech_strain, expected', map(tuple, parametrization_data_stress))
def test_true_stress_scalar(tech_stress, tech_strain, expected):
    np.testing.assert_approx_equal(true_stress(tech_stress, tech_strain), expected, significant=5)
    
    
    

parametrization_fracture_strain= np.array([
    [0.0, 0.0],
    [0.2, 0.22314355131420976],
    [0.3, 0.3566749439387324]
])

@pytest.mark.parametrize('reduction_area_fracture, expected', map(tuple, parametrization_fracture_strain))
def test_true_fracture_strain_scalar(reduction_area_fracture, expected):
    np.testing.assert_approx_equal(true_fracture_strain(reduction_area_fracture), expected, significant=5)


parametrization_fracture_stress= np.array([
    [0.0, 1.0, 0.1, 0],
    [5000.0, 12.0, 0.6, 1041.6666666666665]
   ])

@pytest.mark.parametrize('fracture_force, initial_cross_section, reduction_area_fracture, expected', map(tuple, parametrization_fracture_stress))
def test_true_fracture_stress_scalar(fracture_force, initial_cross_section, reduction_area_fracture, expected):
    np.testing.assert_approx_equal(true_fracture_stress(fracture_force, initial_cross_section, reduction_area_fracture), expected, significant=5)
    


# Copyright (c) 2020-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Gyöngyvér Kiss"
__maintainer__ = __author__

import datetime
import getpass

import numpy as np
import pandas as pd
import h5py
import os

from .exceptions import *
from . import vmap_structures
from .vmap_unit_system import VMAPUnit
from .vmap_element_type import VMAPElementType
from .vmap_attribute import VMAPAttribute
from .vmap_coordinate_system import VMAPCoordinateSystem
from .vmap_section import VMAPSection
from .vmap_metadata import VMAPMetadata
from .vmap_integration_type import VMAPIntegrationType


class VMAPExportError(Exception):
    pass


class VMAPExport:
    """
    The interface class to export a vmap file

    Parameters
    ----------
    file_name : string
        The path to the vmap file to be read

    Raises
    ------
    Exception
        if the file cannot be read an exception is raised.
        So far any exception from the ``h5py`` module is passed through.
    """

    """
    These dictionaries are to provide the data to the SYSTEM datasets. This data is going to come from the import
    """
    _element_types = {
        (2, 3): [0, 'VMAP_ELEM_2D_TRIANGLE_3', 'pyLife 2D 3', 3, 2, -1, -1, -1, -1, -1, [], []],
        (2, 6): [1, 'VMAP_ELEM_2D_TRIANGLE_6', 'pyLife 2D 6', 6, 2, -1, -1, -1, -1, -1, [], []],
        (2, 4): [2, 'VMAP_ELEM_2D_QUAD_4', 'pyLife 2D 4', 4, 2, -1, -1, -1, -1, -1, [], []],
        (2, 8): [3, 'VMAP_ELEM_2D_QUAD_8', 'pyLife 2D 8', 8, 2, -1, -1, -1, -1, -1, [], []],
        (3, 4): [4, 'VMAP_ELEMENT_3D_TETRAHEDRON_4', 'pyLife 3D 4', 4, 3, -1, -1, -1, -1, -1, [], []],
        (3, 10): [5, 'VMAP_ELEMENT_3D_TETRAHEDRON_10', 'pyLife 3D 10', 10, 3, -1, -1, -1, -1, -1, [], []],
        (3, 6): [6, 'VMAP_ELEMENT_3D_WEDGE_6', 'pyLife 3D 6', 6, 3, -1, -1, -1, -1, -1, [], []],
        (3, 15): [7, 'VMAP_ELEMENT_3D_WEDGE_15', 'pyLife 3D 15', 15, 3, -1, -1, -1, -1, -1, [], []],
        (3, 8): [8, 'VMAP_ELEMENT_3D_HEXAHEDRON_8', 'pyLife 3D 8', 8, 3, -1, -1, -1, -1, -1, [], []],
        (3, 20): [9, 'VMAP_ELEMENT_3D_HEXAHEDRON_20', 'pyLife 3D 20', 20, 3, -1, -1, -1, -1, -1, [], []]
    }

    _coordinate_systems = {
        'CARTESIAN': [1, 2, [0.0, 0.0, 0.0], [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0]]
    }

    _sections = {
        'DEFAULT': [0, 'Section-ASSEMBLY_DEFAULT', 1, 0, _coordinate_systems['CARTESIAN'][0], -1, -1]
    }

    _unit_system = {
        'LENGTH': [1, 1.0, 0.0, 'm', 'LENGTH'],
        'MASS': [2, 1.0, 0.0, 'kg', 'MASS'],
        'TIME': [3, 1.0, 0.0, 's', 'TIME'],
        'ELECTRIC_CURRENT': [4, 1.0, 0.0, 'A', 'ELECTRIC CURRENT'],
        'TEMPERATURE': [5, 1.0, 0.0, 'K', 'TEMPERATURE'],
        'AMOUNT_OF_SUBSTANCE': [6, 1.0, 0.0, 'mol', 'AMOUNT OF SUBSTANCE'],
        'LUMINOUS_INTENSITY': [7, 1.0, 0.0, 'cd', 'LUMINOUS INTENSITY'],
    }

    _metadata = {
        'EXPORTER_NAME': ['ExporterName', 'pyLife'],
        'FILE_DATE': ['FileDate', datetime.datetime.now().date().strftime("%Y-%m-%d")],
        'FILE_TIME': ['FileTime', datetime.datetime.now().time().strftime("%H:%M:%S.%f")],
        'DESCRIPTION': ['Description', ''],
        'ANALYSIS_TYPE': ['Analysis Type', ''],
        'USERID': ['User Id', getpass.getuser()]
    }

    def __init__(self, file_name):
        self._file_name = file_name
        try:
            with h5py.File(file_name, 'w') as file:
                self._create_fundamental_groups(file)
            self._dimension = 2
        except OSError:
            if os.path.exists(self._file_name):
                os.remove(self._file_name)
            raise

    @property
    def file_name(self):
        """
        Gets the name of the VMAP file that we are exporting
        """
        return self._file_name

    def variable_column_names(self, parameter_name):
        """
        Gets the column names that the given parameter consists of

        Parameters
        ----------
        parameter_name: string
            The name of the parameter

        Returns
        -------
        The column names of the given parameter in the mesh

        """
        return vmap_structures.column_names[parameter_name][0]

    def variable_location(self, parameter_name):
        """
        Gets the location of the given parameter

        Parameters
        ----------
        parameter_name: string
            The name of the parameter

        Returns
        -------
        The location of the given parameter

        """
        return vmap_structures.column_names[parameter_name][1]

    def set_group_attribute(self, object_path, key, value):
        """
        Sets the 'MYNAME' attribute of the VMAP objects

        Parameters
        ----------
        object_path: string
            The full path to the object that we want to rename
        key: string
            The key of the attribute that we want to set
        value: np.dtype
            The value that we want to set to the attribute

        Returns
        -------
        -
        """
        with h5py.File(self._file_name, 'a') as file:
            try:
                vmap_object = file[object_path]
            except KeyError:
                raise KeyError('VMAP object %s does not exist.' % object_path)
            vmap_object.attrs.create(key, value)

    def add_geometry(self, geometry_name, mesh):
        """
        Exports geometry with given name and mesh data

        Parameters
        ----------
        geometry_name: string
            Name of the geometry to add
        mesh: Pandas DataFrame
            The Data Frame that holds the data of the mesh to export
        Returns
        -------
        self

        """
        with h5py.File(self._file_name, 'a') as file:
            geometry_group = file["/VMAP/GEOMETRY"]
            if geometry_name in geometry_group:
                raise KeyError('Geometry %s already exists' % geometry_name)
            try:
                geometry = self._create_geometry_groups(file, geometry_group, geometry_name)
                self._create_points_datasets(geometry, mesh)
                self._create_elements_dataset(geometry, mesh)
            except Exception as e:
                del geometry_group[geometry_name]
                raise VMAPExportError('An error occurred while creating geometry %s: %s' %
                                      (geometry_name, str(e)))
        return self

    def add_node_set(self, geometry_name, indices, mesh, name=None):
        """
        Exports node-type geometry set into given geometry

        Parameters
        ----------
        geometry_name: string`
            The geometry to where we want to export the geometry set
        indices: Pandas Index
            List of node indices that we want to export
        mesh: Pandas DataFrame
            The Data Frame that holds the data of the mesh to export
        name: value of attribute MYSETNAME

        Returns
        -------
        self

        """
        node_id_set = set(mesh.index.get_level_values('node_id'))
        index_set = set(indices)
        if not index_set.issubset(node_id_set):
            raise KeyError('Provided index set is not a subset of the node indices.')
        self._create_geometry_set(geometry_name, 0, indices, name)
        return self

    def add_element_set(self, geometry_name, indices, mesh, name=None):
        """
        Exports element-type geometry set into given geometry

        Parameters
        ----------
        geometry_name: string
            The geometry to where we want to export the geometry set
        indices: Pandas Index
            List of node indices that we want to export
        mesh: Pandas DataFrame
            The Data Frame that holds the data of the mesh to export
        name: value of attribute MYSETNAME

        Returns
        -------
        self

        """
        element_id_set = set(mesh.index.get_level_values('element_id'))
        index_set = set(indices)
        if not index_set.issubset(element_id_set):
            raise KeyError('Provided index set is not a subset of the element indices.')
        self._create_geometry_set(geometry_name, 1, indices, name)
        return self

    def add_integration_types(self, content):
        """
        Creates system dataset IntegrationTypes with the given content

        Parameters
        ----------
        content: the content of the dataset

        Returns
        -------
        self

        """
        self._create_system_dataset(VMAPIntegrationType, content)
        return self

    def add_variable(self, state_name, geometry_name, variable_name, mesh, column_names=None, location=None):
        """Exports variable into given state and geometry

        Parameters
        ----------
        state_name: string
            State where we want to export the parameter
        geometry_name: string
            Geometry where we want to export the parameter
        variable_name: string
            The name of the variable to export
        mesh: Pandas DataFrame
            The Data Frame that holds the data of the mesh to export
        column_names: List, optional
            The columns that the parameter consists of
        location: Enum, optional
            The location of the parameter
            * 2 - node
            * 3 - element - not supported yet
            * 6 - element nodal

        Returns
        -------
        self

        """
        with h5py.File(self._file_name, 'a') as file:
            try:
                file["VMAP/GEOMETRY/%s" % geometry_name]
            except KeyError:
                raise KeyError("No geometry with the name %s" % geometry_name)

            try:
                state_group = file["/VMAP/VARIABLES/%s" % state_name]
            except KeyError:
                state_group = self._create_group_with_attributes(file['VMAP/VARIABLES'], state_name,
                                                                 VMAPAttribute('MYSTATEINCREMENT', 0),
                                                                 VMAPAttribute('MYSTATENAME', str.encode(state_name,
                                                                                                         'UTF8')),
                                                                 VMAPAttribute('MYSTEPTIME', 0.0),
                                                                 VMAPAttribute('MYTOTALTIME', 0.0))
            try:
                geometry_group = state_group[geometry_name]
            except KeyError:
                geometry_group = self._create_group_with_attributes(state_group, geometry_name,
                                                                    VMAPAttribute('MYSIZE', 0))

            if variable_name in geometry_group:
                raise KeyError("Variable already exists in state %s and geometry %s" % (state_name, geometry_name))

            if column_names is None:
                try:
                    column_names = vmap_structures.column_names[variable_name][0]
                except KeyError:
                    raise KeyError("No column name for variable %s." % variable_name)

            if location is None:
                if variable_name not in vmap_structures.column_names:
                    raise APIUseError(
                        "Need location for unknown variable %s. Please provide one using 'location' parameter."
                        % variable_name)
                location = vmap_structures.column_names[variable_name][1]

            if not isinstance(location, vmap_structures.VariableLocations):
                raise APIUseError("location parameter needs to be of type VariableLocations.")

            try:
                variable_dataset = self._create_group_with_attributes(geometry_group, variable_name,
                                                                      VMAPAttribute('MYCOORDINATESYSTEM', -1),
                                                                      VMAPAttribute('MYDIMENSION', len(column_names)),
                                                                      VMAPAttribute('MYENTITY', 1),
                                                                      VMAPAttribute('MYIDENTIFIER',
                                                                                    len(geometry_group)),
                                                                      VMAPAttribute('MYINCREMENTVALUE', 1),
                                                                      VMAPAttribute('MYLOCATION', location.value),
                                                                      VMAPAttribute('MYMULTIPLICITY', 1),
                                                                      VMAPAttribute('MYTIMEVALUE', 0.0),
                                                                      VMAPAttribute('MYUNIT', -1),
                                                                      VMAPAttribute('MYVARIABLEDEPENDENCY', b' '),
                                                                      VMAPAttribute('MYVARIABLEDESCRIPTION',
                                                                                    str.encode(
                                                                                        'pyLife: %s' % variable_name,
                                                                                        'UTF8')),
                                                                      VMAPAttribute('MYVARIABLENAME',
                                                                                    str.encode(variable_name, 'UTF8')))
                if location == vmap_structures.VariableLocations.NODE:
                    node_ids_info = mesh.groupby('node_id').first()
                    variable_dataset.create_dataset('MYGEOMETRYIDS', data=np.array([node_ids_info.index]).T,
                                                    dtype=np.int32, chunks=True)
                    variable_dataset.create_dataset('MYVALUES', data=node_ids_info[column_names].values, chunks=True)
                else:
                    element_ids = mesh.index.get_level_values('element_id').drop_duplicates().values
                    variable_dataset.create_dataset('MYGEOMETRYIDS', data=np.array([element_ids]).T, dtype=np.int32,
                                                    chunks=True)
                    variable_dataset.create_dataset('MYVALUES', data=mesh[column_names], chunks=True)

                geometry_group.attrs['MYSIZE'] = geometry_group.attrs['MYSIZE'] + 1
            except Exception as e:
                del geometry_group[variable_name]
                raise VMAPExportError('An error occurred while creating variable %s: %s'
                                      % (variable_name, str(e)))
        return self

    def _create_group_with_attributes(self, parent_group, group_name, *args):
        group = parent_group.create_group(group_name)
        if args is not None:
            for attr in args:
                group.attrs.create(attr.name, attr.value)
        return group

    def _create_compound_attribute(self, parent, attr_name, field_names, field_types, field_values):
        dt = np.dtype({"names": field_names, "formats": field_types})
        compound_attribute = np.array([field_values], dt)
        parent.attrs.create(attr_name, compound_attribute)

    def _create_fundamental_groups(self, file):
        vmap_group = file.create_group('VMAP')
        self._create_compound_attribute(vmap_group, 'VERSION', ["myMajor", "myMinor", "myPatch"],
                                        ['<i4', '<i4', '<i4'], ('0', '5', '2'))

        self._create_group_with_attributes(vmap_group, 'GEOMETRY')
        self._create_group_with_attributes(vmap_group, 'MATERIAL')
        self._create_group_with_attributes(vmap_group, 'SYSTEM')

        self._create_system_dataset(VMAPCoordinateSystem, self._coordinate_systems)
        self._create_system_dataset(VMAPElementType, self._element_types)
        self._create_system_dataset(VMAPMetadata, self._metadata)
        self._create_system_dataset(VMAPSection, self._sections)
        self._create_system_dataset(VMAPUnit, self._unit_system)

        self._create_group_with_attributes(vmap_group, 'VARIABLES')

    def _create_system_dataset(self, class_name, dataset_content):
        attribute_list = []
        for content_key in dataset_content:
            dataset = class_name(*dataset_content[content_key])
            attribute_list.append(dataset.attributes)
        name = dataset.dataset_name
        dt_type = dataset.dtype
        path = dataset.group_path
        compound_dataset = dataset.compound_dataset
        if compound_dataset:
            d = np.array([attribute_list], dtype=dt_type).T
            chunked = True
        else:
            d = np.array(attribute_list, dtype=dt_type)
            chunked = None
        with h5py.File(self._file_name, 'a') as file:
            system_group = file[path]
            if name in system_group:
                raise KeyError('Dataset %s already exists in SYSTEM')

            try:
                system_group.create_dataset(name, dtype=dt_type, data=d, chunks=chunked)
            except Exception as e:
                del system_group[name]
                raise VMAPExportError('An error occurred while creating dataset %s: %s' %
                                      (name, str(e)))
        return self

    def _create_geometry_groups(self, file, geometry_group, geometry_name):
        geometry = self._create_group_with_attributes(geometry_group, geometry_name)
        self._create_group_with_attributes(geometry, 'ELEMENTS', VMAPAttribute('MYSIZE', np.int64(0)))
        self._create_group_with_attributes(geometry, 'GEOMETRYSETS', VMAPAttribute('MYSIZE', 0))
        self._create_group_with_attributes(geometry, 'POINTS',
                                           VMAPAttribute('MYSIZE', np.int64(0)),
                                           VMAPAttribute('MYCOORDINATESYSTEM',
                                                         self._coordinate_systems['CARTESIAN'][0]))
        return geometry

    def _create_elements_dataset(self, geometry, mesh):
        dt_type = np.dtype({"names": ["myIdentifier", "myElementType", "myCoordinateSystem",
                                      "myMaterialType", "mySectionType", "myConnectivity"],
                            "formats": ['<i4', '<i4', '<i4', '<i4', '<i4', h5py.special_dtype(vlen=np.dtype('int32'))]})
        element_connectivities = mesh.groupby('element_id')
        element_ids = []
        node_ids_list = []
        element_types_list = []
        coordinate_system = np.empty(len(element_connectivities), dtype=np.int32)
        coordinate_system.fill(1)
        material_type = np.empty(len(element_connectivities), dtype=np.int32)
        material_type.fill(0)
        section_type = np.empty(len(element_connectivities), dtype=np.int32)
        section_type.fill(self._sections['DEFAULT'][0])

        for element_connectivity in element_connectivities:
            c = element_connectivity[1].index.get_level_values('node_id').values
            node_ids_list.append(c)
            element_ids.append(element_connectivity[0])
            element_type = self._element_types[self._dimension, c.size][0]
            element_types_list.append(element_type)

        element_types = np.asarray(element_types_list)
        connectivity = np.asarray(node_ids_list)
        d = np.array([list(zip(element_ids, element_types, coordinate_system,
                               material_type, section_type, connectivity))], dtype=dt_type).T
        elements_group = geometry['ELEMENTS']
        elements_group.create_dataset("MYELEMENTS", dtype=dt_type, data=d, chunks=True)
        elements_group.attrs['MYSIZE'] = np.int64(len(element_ids))

    def _create_points_datasets(self, geometry, mesh):
        node_ids_info = mesh.groupby('node_id').first()
        points_group = geometry['POINTS']
        points_group.create_dataset('MYIDENTIFIERS', data=np.reshape(node_ids_info.index, (-1, 1)), dtype=np.int32,
                                    chunks=True)
        if 'z' in node_ids_info:
            z = node_ids_info['z'].to_numpy()
            if not (z[0] == z).all():
                self._dimension = 3
            points_group.create_dataset('MYCOORDINATES', data=node_ids_info[['x', 'y', 'z']].values, chunks=True)
        else:
            points_group.create_dataset('MYCOORDINATES', data=node_ids_info[['x', 'y']].values, chunks=True)
        points_group.attrs['MYSIZE'] = np.int64(node_ids_info.index.size)

    def _create_geometry_set(self, geometry_name, object_type, indices, name=None):
        if name is None:
            name = ''

        if not isinstance(name, str):
            raise TypeError("Invalid set name (must be a string).")

        with h5py.File(self._file_name, 'a') as file:
            try:
                geometry_set_group = file["VMAP/GEOMETRY/%s/GEOMETRYSETS" % geometry_name]
            except KeyError:
                raise KeyError("No geometry with the name %s" % geometry_name)
            try:
                set_size = geometry_set_group.attrs['MYSIZE']
                geometry_set_name = str(f"{set_size:06d}")
                geometry_set = self._create_group_with_attributes(geometry_set_group, geometry_set_name,
                                                                  VMAPAttribute('MYIDENTIFIER', set_size),
                                                                  VMAPAttribute('MYSETINDEXTYPE', 1),
                                                                  VMAPAttribute('MYSETNAME', str.encode(name, 'UTF-8')),
                                                                  VMAPAttribute('MYSETTYPE', object_type))
                geometry_set.create_dataset('MYGEOMETRYSETDATA', data=pd.DataFrame(indices),
                                            dtype=np.int32, chunks=True)
                geometry_set_group.attrs['MYSIZE'] = set_size + 1
            except Exception as e:  # pragma: no cover (all possible exceptions should have been caught already)
                del geometry_set_group[geometry_set_name]
                raise VMAPExportError(
                    "An error occurred while creating geometry set %s in geometry %s: %s"
                    "If this is reproducible, please file a bug report."
                    % (geometry_set_name, geometry_name, str(e))
                )



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Implementation of the miner rule for fatigue analysis
=====================================================

Currently, the following implementations are part of this module:

* Miner Elementary
* Miner Haibach

References
----------
* M. Wächter, C. Müller and A. Esderts, "Angewandter Festigkeitsnachweis nach {FKM}-Richtlinie"
  Springer Fachmedien Wiesbaden 2017, https://doi.org/10.1007/978-3-658-17459-0

* E. Haibach, "Betriebsfestigkeit", Springer-Verlag 2006, https://doi.org/10.1007/3-540-29364-7
"""

__author__ = "Cedric Philip Wagner"
__maintainer__ = "Johannes Mueller"

from abc import ABC, abstractmethod

import numpy as np
import pandas as pd

from pylife.strength.fatigue import Fatigue
from pylife.materiallaws.woehlercurve import WoehlerCurve

import pylife.strength.solidity as SOL


class MinerBase(WoehlerCurve, ABC):
    """Basic functions related to miner-rule (original).

    Uses the constructor of :class:`~pylife.materiallaws.WoehlerCurve`.
    """

    def finite_life_factor(self, N):
        """Calculate *finite life factor* according to Waechter2017 (p. 96).

        Parameters
        ----------
        N : float
            Collective range (sum of cycle numbers) of load collective
        """
        return np.power(self.ND/N, 1./self.k_1)

    def effective_damage_sum(self, collective):
        """Compute *effective damage sum* D_m.

        Refers to the formula given in Waechter2017, p. 99

        Parameters
        ----------
        collective : a load collective
            the multiple of the lifetime

        Returns
        -------
        effective_damage_sum : float or :class:`pandas.Series`
            The effective damage sums for the collective
        """
        A = self.lifetime_multiple(collective)
        return effective_damage_sum(A)

    def gassner_cycles(self, collective):
        """Compute the cycles of the Gassner line for a certain load collective.

        Parameters
        ----------
        collective : :class:`~pylife.stress.rainflow.LoadCollective` or similar
            The load collective

        Returns
        -------
        cycles
            The cycles for the collective

        Note
        ----
        The absolute load levels of the collective are important.
        """
        return self.cycles(collective.amplitude.max()) * self.lifetime_multiple(collective)

    @abstractmethod
    def lifetime_multiple(self, collective):
        """Compute the lifetime multiple according to the corresponding Miner rule.

        Needs to be implemented in the class implementing the Miner rule.

        Parameters
        ----------
        collective : :class:`~pylife.stress.rainflow.LoadCollective` or similar
            The load collective

        Returns
        -------
        lifetime_multiple : float > 0
            lifetime multiple
        """

        pass


@pd.api.extensions.register_series_accessor('gassner_miner_elementary')
class MinerElementary(MinerBase):
    """Implementation of Miner Elementary according to Waechter2017."""

    def gassner(self, collective):
        """Calculate the Gaßner shift according to Miner Elementary.

        Parameters
        ----------
        collective : :class:`~pylife.stress.rainflow.LoadCollective` or similar
            The load collective

        Returns
        -------
        gassner : :class:`~pylife.stength.Fatigue`
            The Gaßner shifted fatigue strength object.
        """
        gassner = self.to_pandas().copy()
        gassner['ND'] = self.ND * self.lifetime_multiple(collective)
        return Fatigue(gassner)

    def lifetime_multiple(self, collective):
        """Compute the lifetime multiple according to Miner Elementary.

        Described in Waechter2017 as "Lebensdauervielfaches, A_ele".

        Parameters
        ----------
        collective : :class:`~pylife.stress.rainflow.LoadCollective` or similar
            The load collective

        Returns
        -------
        lifetime_multiple : float > 0
            lifetime multiple
        """
        return 1. / SOL.haibach(collective, self.k_1)


@pd.api.extensions.register_series_accessor('gassner_miner_haibach')
class MinerHaibach(MinerBase):
    """Miner-modified according to Haibach (2006).

    Warnings
    --------
    Contrary to Miner Elementary, the lifetime multiple is not constant but
    dependent on the evaluated load level!  That is why there is no method for
    the Gaßner shift.
    """

    def lifetime_multiple(self, collective):
        """Compute the lifetime multiple for Miner-modified according to Haibach.

        Refer to Haibach (2006), p. 291 (3.21-61). The lifetime multiple can be
        expressed in respect to the maximum amplitude so that
        N_lifetime = N_Smax * A

        Parameters
        ----------
        collective : :class:`~pylife.stress.rainflow.LoadCollective` or similar
            The load collective

        Returns
        -------
        lifetime_multiple : float > 0
            lifetime multiple
            return value is 'inf' if maximum collective amplitude < SD
        """
        s_a = collective.amplitude
        max_amp = s_a.max()

        cycles = collective.cycles

        s_a = s_a / max_amp
        x_D = self.SD / max_amp

        i_full_damage = (s_a >= x_D)
        i_reduced_damage = (s_a < x_D)

        s_full_damage = s_a[i_full_damage]
        s_reduced_damage = s_a[i_reduced_damage]

        n_full_damage = cycles[i_full_damage]
        n_reduced_damage = cycles[i_reduced_damage]

        # first expression of the summation term in the denominator
        sum_1 = np.dot(n_full_damage, (s_full_damage**self.k_1))
        sum_2 = x_D**(1 - self.k_1) * np.dot(n_reduced_damage, (s_reduced_damage**(2 * self.k_1 - 1)))

        return cycles.sum() / (sum_1 + sum_2)


def effective_damage_sum(lifetime_multiple):
    """Compute *effective damage sum*.

    Refers to the formula given in Waechter2017, p. 99

    Parameters
    ----------
    A : float or np.ndarray (with 1 element)
        the multiple of the lifetime

    Returns
    d_m : float
        the effective damage sum
    """
    d_min = 0.3  # minimum as suggested by FKM
    d_max = 1.0

    d_m_no_limits = 2. / (lifetime_multiple**(1./4.))
    d_m = min(
        max(d_min, d_m_no_limits),
        d_max
    )

    return d_m



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import numpy as np
import pandas as pd
from scipy import signal as sg
from pylife.stress import timesignal as pts


def create_input_DF():
    fs = 2048
    t = np.arange(0, 30, 1/fs)
    ts_df = pd.DataFrame({"sin1": np.sin(10* 2 * np.pi * t),
                          "sin2": 2 * np.sin(10* 2 * np.pi * t + 0.05),
                          "cos": 10 * np.cos(5* 2 * np.pi * t),
                          "wn": np.random.rand(len(t))},
                         index=t)
    ts_df.index.name = "t"
    return ts_df

def test_fs_calc_true():
    df = create_input_DF()
    fs = 2048
    assert fs == pts.fs_calc(df)

def test_fs_calc_false():
    df = create_input_DF()
    df.index = df.index.astype(str)
    assert 1 == pts.fs_calc(df)

def test_resample_acc_sine():
    # sine
    omega = 10*2*np.pi  # Hz
    ts_sin = pd.DataFrame(np.sin(
        omega*np.arange(0, 2, 1/1024)), index=np.arange(0, 2, 1/1024))
    expected_sin = ts_sin.describe().drop(['count', 'mean', '50%', '25%', '75%'])
    test_sin = pts.resample_acc(ts_sin, int(12*omega)).describe().drop(
        ['count', 'mean', '50%', '25%', '75%'])
    pd.testing.assert_frame_equal(test_sin, expected_sin, rtol=2)

def test_resample_acc_wn():
    # white noise
    ts_wn = pd.DataFrame(np.random.randn(129), index=np.linspace(0, 1, 129))
    expected_wn = ts_wn.describe()
    test_wn = pts.resample_acc(ts_wn, fs = pts.fs_calc(ts_wn)).describe()
    pd.testing.assert_frame_equal(test_wn, expected_wn, check_exact=True)
def test_resample_acc_Sor():
    # SoR
    t = np.arange(0, 20, 1/4096)
    ts_sin = pd.DataFrame(np.sin(10 * 2 * np.pi * t), index=t)
    expected_sin = ts_sin.describe().drop(
        ['count', 'mean', '50%', '25%', '75%'])
    test_sin = pts.resample_acc(ts_sin, 2048).describe().drop(
        ['count', 'mean', '50%', '25%', '75%'])
    pd.testing.assert_frame_equal(test_sin, expected_sin, rtol=1e-2)
def test_resample_acc_sawtooth():
    # sawtooth
    t=t = np.arange(0, 10, 1/4096)
    ts_st = pd.DataFrame(sg.sawtooth(2 * np.pi * 1 * t), index=t)
    expected_st = ts_st.describe().drop(['count', '50%', 'mean'])
    test_st = pts.resample_acc(ts_st, 1024).describe().drop(['count', '50%', 'mean'])
    pd.testing.assert_frame_equal(test_st, expected_st, rtol=1e-2)

def test_ps_df():
    sample_frequency = 512
    t = np.linspace(0, 60, 60 * sample_frequency)
    ts_df = pd.DataFrame(data = np.array([np.random.randn(len(t)), np.sin(2 * np.pi * 10 * t)]).T,
                         columns=["wn", "sine"],
                         index=t)
    test_psd = pts.psd_df(ts_df, 512)

    np.testing.assert_allclose(test_psd.sum().values, np.array([1, 0.5]), rtol=1e-1)

# def test_running_stats_filt():
#     t = np.linspace(0, 1, 2048+1)
#     sin1 = -abs(150*np.sin(2*np.pi*10*t)+152)
#     sin2 = 300*np.sin(2*np.pi*10*t)
#     wn1 = np.ones(len(t))
#     wn1[500] = -250
#     wn2 = np.zeros(len(t))
#     wn2[0] = 350
#     df = pd.DataFrame(np.hstack((
#         sin1,
#         sin2,
#         wn1,
#         wn2)), columns=['data'])
#     df_sep = pd.DataFrame(np.vstack((
#         sin1,
#         sin2,
#         wn1,
#         wn2))).T
#     df_sep.columns = ['sin1', 'sin2', 'wn1', 'wn2']

#     test_rms = pts.TimeSignalPrep(df).running_stats_filt(
#         col='data', window_length=2049, buffer_overlap=0.0, limit=0.95, method="rms")
#     test_rms.columns = ['sin2']
#     pd.testing.assert_frame_equal(test_rms, pd.DataFrame(df_sep['sin2']),
#                                   check_index_type=False)

#     test_max = pts.TimeSignalPrep(df).running_stats_filt(
#         col='data', window_length=2049, buffer_overlap=0.0, limit=0.95, method="max")
#     test_max.columns = ['wn2']
#     pd.testing.assert_frame_equal(test_max, pd.DataFrame(df_sep['wn2']),
#                                   check_index_type=False)

#     test_min = pts.TimeSignalPrep(df).running_stats_filt(
#         col='data', window_length=2049, buffer_overlap=0.0, limit=1, method="min")
#     test_min.columns = ['sin1']
#     pd.testing.assert_frame_equal(test_min, pd.DataFrame(df_sep['sin1']),
#                                   check_index_type=False)

#     test_abs = pts.TimeSignalPrep(df).running_stats_filt(
#         col='data', window_length=2049, buffer_overlap=0.0, limit=0.1, method="abs")
#     pd.testing.assert_frame_equal(test_abs, df,
#                                   check_index_type=False)
#     test_sor = pts.resample_acc(df, 2048).describe().drop(
#         ['count', 'mean', '50%', '25%', '75%'])
#     # return expected_sor, test_sor
#     pd.testing.assert_frame_equal(test_sor, expected_sor, rtol=1e-2)
#     return


# %%
# Test timeseries

@pytest.fixture
def ts_inp():
    size_df = 10
    return pd.DataFrame(
        index=np.arange(size_df)+1,
        data={'x': np.arange(size_df)**2}
    )


@pytest.fixture
def ts_prep_rolling(ts_inp):
    return pts._prepare_rolling(ts_inp)


@pytest.fixture
def df_rolled(ts_prep_rolling):
    return pts._roll_dataset(ts_prep_rolling, window_size=3, overlap=1)


@pytest.fixture
def extracted_features(df_rolled):
    return pts._extract_feature_df(df_rolled, feature="maximum")


@pytest.fixture
def grid_points(ts_prep_rolling, extracted_features):
    return pts._select_relevant_windows(ts_prep_rolling, extracted_features, 'x__maximum',
                                        fraction_max=0.24, window_size=3,
                                        overlap=1, n_gridpoints=3)


@pytest.fixture
def poly_gridpoints(grid_points, ts_prep_rolling):
    ts_time = ts_prep_rolling.copy()
    delta_t = ts_time.index[1]-ts_time.index[0]
    line = pd.DataFrame(ts_time.iloc[:1], index=[- delta_t])
    ts_time = pd.concat([ts_time, line], ignore_index=False).sort_index()
    ts_time.index = ts_time.index + delta_t
    ts_time['time'] = ts_time.index.values

    return pts._polyfit_gridpoints(
        grid_points, ts_time, order=1, verbose=False, n_gridpoints=3
    )


@pytest.fixture
def ts_cleaned(ts_inp):
    return pts.clean_timeseries(
        ts_inp, 'x', window_size=3,
        overlap=1, feature="maximum", n_gridpoints=2,
        percentage_max=0.24, order=1
    )


def test_prepare_rolling(ts_inp, ts_prep_rolling):
    size_df = 10
    exact_res = pd.DataFrame({
        'x': np.arange(size_df)**2,
        'id': 0,
        'time': np.int64(np.arange(size_df))
    })
    exact_res.index = exact_res["time"]

    pd.testing.assert_frame_equal(exact_res, ts_prep_rolling, check_dtype=False, check_index_type=False)


def test_roll_dataset(df_rolled):
    x = np.array([0, 1, 2, 2, 3, 4, 4, 5, 6, 6, 7, 8])

    df = pd.DataFrame({
        'id': np.int64(np.zeros(len(x), dtype=int)),
        'max_time': [2, 2, 2, 4, 4, 4, 6, 6, 6, 8, 8, 8]
    })

    exact_res = pd.DataFrame({'x': x**2}, index=np.arange(len(x)))

    exact_res['id'] = pd.MultiIndex.from_frame(df)
    exact_res['time'] = np.int64(x)

    pd.testing.assert_frame_equal(exact_res, df_rolled, check_dtype=False, check_index_type=False)


# %%

@pytest.mark.skipif(not pts._HAVE_TSFRESH, reason="Don't have tsfresh")
def test_extract_feature_df(extracted_features):
    exact_res = pd.DataFrame({'x__maximum': [4., 16., 36., 64.]})
    pd.testing.assert_frame_equal(exact_res, extracted_features)


# %%


@pytest.mark.skipif(not pts._HAVE_TSFRESH, reason="Don't have tsfresh")
def test_select_relevant_windows(ts_prep_rolling, grid_points):
    ts_prep_selected = ts_prep_rolling.copy()
    ts_prep_selected.iloc[0:3, 0] = None

    exact_res = ts_prep_selected
    pd.testing.assert_frame_equal(exact_res, grid_points)


# %%


@pytest.mark.skipif(not pts._HAVE_TSFRESH, reason="Don't have tsfresh")
def test_polyfit_gridpoints1(grid_points, poly_gridpoints):
    x = [0, 4]
    y = [0, 9]
    z = np.polyfit(x, y, 1)
    p = np.poly1d(z)

    delta_t = grid_points.index[1]-grid_points.index[0]
    line = pd.DataFrame(grid_points.iloc[:1], index=[- delta_t])
    grid_points = pd.concat([grid_points, line], ignore_index=False).sort_index()
    grid_points["time"] = grid_points.index + delta_t
    grid_points.index = grid_points["time"]
    grid_points.iloc[0, :] = 0

    grid_points.iloc[1, 0] = p(1)
    grid_points.iloc[2, 0] = p(2)
    grid_points.iloc[3, 0] = p(3)
    exact_res = grid_points
    exact_res["time"] = exact_res.index.values
    pd.testing.assert_frame_equal(exact_res, poly_gridpoints)


# %%

@pytest.mark.skipif(not pts._HAVE_TSFRESH, reason="Don't have tsfresh")
def test_clean_timeseries(poly_gridpoints, ts_cleaned):
    poly_gridpoints.pop("id")
    poly_gridpoints.index = poly_gridpoints["time"]
    pd.testing.assert_frame_equal(poly_gridpoints, ts_cleaned)


# %% Test timeseries 2

@pytest.fixture
def ts_gaps():
    t = np.linspace(0, np.pi, 20)
    y1 = np.sin(4*t)+1
    y2 = np.sin(t)+1
    return pd.DataFrame(index=t,
                        data=np.array([y1, y2]).T)


@pytest.fixture
def df_gaps_prep(ts_gaps):
    df_gaps_prep = pts._prepare_rolling(ts_gaps)
    return df_gaps_prep


@pytest.fixture
def df_gaps_rolled(df_gaps_prep):
    df_gaps_rolled = pts._roll_dataset(df_gaps_prep, window_size=5, overlap=2)
    return df_gaps_rolled


@pytest.fixture
def extraced_feature_gaps(df_gaps_rolled):
    extraced_feature_gaps = pts._extract_feature_df(
        df_gaps_rolled, feature="maximum")
    return extraced_feature_gaps


@pytest.fixture
def grid_points_gaps(df_gaps_prep, extraced_feature_gaps):
    return pts._select_relevant_windows(
        df_gaps_prep, extraced_feature_gaps,
        "0__maximum", fraction_max=0.80,
        window_size=5, overlap=2
    )


@pytest.mark.skipif(not pts._HAVE_TSFRESH, reason="Don't have tsfresh")
def test_select_relevant_windows2(df_gaps_prep, grid_points_gaps):
    ts_gaps_selected = df_gaps_prep.copy()
    ts_gaps_selected.iloc[3*2:3*2+5, 0:2] = np.nan
    ts_gaps_selected.iloc[15:20, 0:2] = np.nan
    list = [ts_gaps_selected.index[6], ts_gaps_selected.index[7],
            ts_gaps_selected.index[15], ts_gaps_selected.index[16]]

    exact_res = ts_gaps_selected.drop(list, axis=0)
    pd.testing.assert_frame_equal(exact_res, grid_points_gaps)


# %%

@pytest.mark.skipif(not pts._HAVE_TSFRESH, reason="Don't have tsfresh")
def test_clean_timeseries2(ts_gaps, df_gaps_prep, grid_points_gaps):
    poly_gridpoints_gaps = pts._polyfit_gridpoints(
        grid_points_gaps, df_gaps_prep, order=3, verbose=False, n_gridpoints=3)
    poly_gridpoints_gaps = poly_gridpoints_gaps.dropna(axis=0, how='any')
    poly_gridpoints_gaps.pop("id")
    exact_res = poly_gridpoints_gaps
    ts_cleaned = pts.clean_timeseries(ts_gaps, "0", window_size=5,
                                       overlap=2, feature="maximum", n_gridpoints=3,
                                       percentage_max=0.91, order=3)
    pd.testing.assert_frame_equal(exact_res, ts_cleaned)

# %% Test timeseries 3


@pytest.mark.skipif(not pts._HAVE_TSFRESH, reason="Don't have tsfresh")
def test_clean_timeseries3():
    t1 = np.linspace(0, 2*np.pi, 1000, endpoint=False)
    t2 = np.linspace(2*np.pi, 4*np.pi, 1000)
    y1 = 100 * np.cos(8*t1)
    y2 = 0.1 * np.sin(8*t2)
    t = np.append(t1, t2)
    y = np.append(y1, y2)
    ts_sin = pd.DataFrame(index=t,
                          data=y)

    ts_sin_test = ts_sin.copy()
    ts_cleaned = pts.clean_timeseries(ts_sin_test, "0", window_size=100,
                                       overlap=99, feature="maximum", n_gridpoints=5,
                                       percentage_max=0.3, order=3)

    ts_time = pts._prepare_rolling(ts_sin)

    exact_res = pd.DataFrame(index=t1, data=y1)
    exact_res = pts._prepare_rolling(exact_res)

    delta_t = exact_res.index[1]-exact_res.index[0]
    line = pd.DataFrame(exact_res.iloc[:1], index=[- delta_t])
    exact_res = pd.concat([exact_res, line], ignore_index=False).sort_index()
    exact_res.iloc[0, :] = 0
    exact_res["time"] = exact_res.index + delta_t
    exact_res.index = exact_res["time"]
    exact_res.pop('id')

    pd.testing.assert_frame_equal(exact_res, ts_cleaned)



# -*- coding: utf-8 -*-
"""
Created on Wed Sep  8 14:28:26 2021

@author: KRD2RNG
"""
import numpy as np
import matplotlib.pyplot as plt
import matplotlib

def plot_rf(rf_series_dict):
    fig = plt.figure(figsize=(12, 20))
    cmap = matplotlib.colormaps.get_cmap('jet')
    for ii, key in enumerate(rf_series_dict.keys()):
        rf_series = rf_series_dict[key]
        ax = fig.add_subplot(1,len(rf_series_dict),ii + 1, projection='3d')

        froms = rf_series.index.get_level_values("from").mid
        tos = rf_series.index.get_level_values("to").mid

        width = rf_series.index.get_level_values('from').length.min()
        depth = rf_series.index.get_level_values('to').length.min()
        bottom = np.zeros_like(rf_series)

        max_height = np.max(rf_series)
        min_height = np.min(rf_series)
        rgba = [cmap((k-min_height)/max_height) for k in rf_series]
        ax.set_xlabel('From')
        ax.set_ylabel('To')
        ax.set_zlabel('Count')
        ax.bar3d(froms.ravel(), tos.ravel(), bottom, width, depth, rf_series,
                 color=rgba, shade=True, zsort='average')
        ax.set_title(key)
    return fig



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

import pandas as pd
import numpy as np
import numpy.testing as testing

import pylife.mesh.meshmapping


def points_2d(flen, tlen):
    fx = np.arange(flen, dtype=np.float64)
    fy = np.arange(flen, dtype=np.float64)

    fxx, fyy = np.meshgrid(fx, fy, indexing='ij')
    from_vals = (fyy*2.).flatten()

    tx = np.arange(tlen, dtype=np.float64)
    ty = np.arange(tlen, dtype=np.float64) + 0.5
    txx, tyy = np.meshgrid(tx, ty, indexing='ij')

    fpoints = np.vstack((fxx.flatten(), fyy.flatten())).T
    tpoints = np.vstack((txx.flatten(), tyy.flatten())).T

    from_df = pd.DataFrame(fpoints, columns=['x', 'y'])
    from_df['value'] = from_vals

    to_df = pd.DataFrame(tpoints,  columns=['x', 'y'])

    return from_df, to_df


def points_3d(flen, tlen):
    fx = np.arange(flen, dtype=np.float64)
    fy = np.arange(flen, dtype=np.float64)
    fz = np.arange(flen, dtype=np.float64)

    fxx, fyy, fzz = np.meshgrid(fx, fy, fz, indexing='ij')
    from_vals = (fyy*2.).flatten()

    tx = np.arange(tlen, dtype=np.float64)
    ty = np.arange(tlen, dtype=np.float64) + 0.5
    tz = np.arange(tlen, dtype=np.float64)
    txx, tyy, tzz = np.meshgrid(tx, ty, tz, indexing='ij')

    fpoints = np.vstack((fxx.flatten(), fyy.flatten(), fzz.flatten())).T
    tpoints = np.vstack((txx.flatten(), tyy.flatten(), tzz.flatten())).T

    from_df = pd.DataFrame(fpoints, columns=['x', 'y', 'z'])
    from_df['value'] = from_vals

    to_df = pd.DataFrame(tpoints,  columns=['x', 'y', 'z'])

    return from_df, to_df


def test_mapping_2d_linear():
    flen = 241
    tlen = 21

    from_df, to_df = points_2d(flen, tlen)
    mapper = to_df.meshmapper
    to_vals = mapper.process(from_df, 'value')['value'].to_numpy()
    from_vals = from_df['value'].to_numpy()

    for x in range(tlen):
        for y in range(tlen):
            assert to_vals[x*tlen+y] == (from_vals[x*flen+y] + from_vals[x*flen+y+1])/2.


def test_mapping_2d_linear_out_of_map():
    flen = 241
    tlen = 21

    from_df, to_df = points_2d(flen, tlen)
    mapper = to_df.meshmapper
    from_vals = from_df['value'].to_numpy()

    with pytest.raises(Exception):
        mapper.process(from_vals)


def test_mapping_2d_cubic():
    flen = 241
    tlen = 21

    from_df, to_df = points_2d(flen, tlen)
    mapper = to_df.meshmapper
    to_vals = mapper.process(from_df, 'value', method='cubic')['value'].to_numpy()
    from_vals = from_df['value'].to_numpy()

    for x in range(tlen):
        for y in range(tlen):
            testing.assert_almost_equal(to_vals[x*tlen+y], (from_vals[x*flen+y] + from_vals[x*flen+y+1])/2.)


def test_mapping_3d_linear():
    flen = 23
    tlen = 11

    from_df, to_df = points_3d(flen, tlen)
    mapper = to_df.meshmapper
    to_vals = mapper.process(from_df, 'value')['value'].to_numpy()
    from_vals = from_df['value'].to_numpy()

    for x in range(tlen):
        for y in range(tlen):
            for z in range(tlen):
                assert (to_vals[x*tlen*tlen+y*tlen+z]
                        == (from_vals[x*flen*flen+y*flen+z] + from_vals[x*flen*flen+(y+1)*flen+z])/2.)


def test_mapping_3d_cubic():
    flen = 23
    tlen = 11

    from_df, to_df = points_3d(flen, tlen)
    mapper = to_df.meshmapper

    with pytest.raises(ValueError, match=r'.*cubic.*'):
        mapper.process(from_df, 'value', method='cubic')



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np
import pytest

import pylife.utils.probability_data as PD
from pylife.utils.functions import rossow_cumfreqs


@pytest.fixture
def pd():
    occurrences = [0.19144096, 2.32066184, 0.07425225, 1.32781569, 0.77430177,
                   4.65758492, 9.35734404, 0.72244567, 0.29089061, 1.57460074] # randomly generated
    probs = rossow_cumfreqs(len(occurrences))
    return PD.ProbabilityFit(probs, np.sort(occurrences))

def test_probability_fit_inconsistent():
    with pytest.raises(ValueError, match="probs and occurrence arrays must have the same 1D shape."):
        _ = PD.ProbabilityFit([1, 2], [1, 3, 4])

def test_probability_fit_occurrences():
    pd = PD.ProbabilityFit([1, 2, 4], [1, 3, 4])
    np.testing.assert_array_equal(pd.occurrences, np.array([1., 3., 4.]))

def test_probability_fit_slope(pd):
    np.testing.assert_approx_equal(pd.slope, 1.440756470107515)

def test_probability_fit_intecept(pd):
    np.testing.assert_approx_equal(pd.intercept, 0.04474656548145248)

def test_precentiles(pd):
    expected = np.array([-1.517929, -0.989169, -0.649324, -0.372289, -0.121587,
                         0.121587, 0.372289,  0.649324,  0.989169,  1.517929])
    np.testing.assert_allclose(pd.percentiles, expected, rtol=1e-4)

def test_insufficient_data():
    occurrences = [100.0]
    probs = [-1.1]
    with pytest.raises(ValueError, match=r'Need at least two datapoints for probabilities and occurrences.'):
        _ = PD.ProbabilityFit(probs, occurrences)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''Helper to process mesh based data

Data that is distributed over a geometrical body, e.g. a stress tensor
distribution on a component, is usually transported via a mesh. The
meshes are a list of items (e.g. nodes or elements of a FEM mesh),
each being described by the geometrical coordinates and the local data
values, like for example the local stress tensor data.

In a plain mesh (see :class:`PlainMesh`) there is no further
relation between the items is known, whereas a complete FEM mesh (see
:class:`Mesh`) there is also information on the connectivity
of the nodes and elements.

Examples
--------
Read in a mesh from a vmap file:


>>> df = (vm = pylife.vmap.VMAPImport('demos/plate_with_hole.vmap')
             .make_mesh('1', 'STATE-2')
             .join_variable('STRESS_CAUCHY')
             .join_variable('DISPLACEMENT')
             .to_frame())
>>> df.head()
                            x         y    z        S11       S22  S33        S12  S13  S23        dx        dy   dz
element_id node_id
1          1734     14.897208  5.269875  0.0  27.080811  6.927080  0.0 -13.687358  0.0  0.0  0.005345  0.000015  0.0
           1582     14.555333  5.355806  0.0  28.319006  1.178649  0.0 -10.732705  0.0  0.0  0.005285  0.000003  0.0
           1596     14.630658  4.908741  0.0  47.701195  5.512213  0.0 -17.866833  0.0  0.0  0.005376  0.000019  0.0
           4923     14.726271  5.312840  0.0  27.699907  4.052865  0.0 -12.210032  0.0  0.0  0.005315  0.000009  0.0
           4924     14.592996  5.132274  0.0  38.010101  3.345431  0.0 -14.299768  0.0  0.0  0.005326  0.000013  0.0

Get the coordinates of the mesh.

>>> df.plain_mesh.coordinates.head()
                            x         y    z
element_id node_id
1          1734     14.897208  5.269875  0.0
           1582     14.555333  5.355806  0.0
           1596     14.630658  4.908741  0.0
           4923     14.726271  5.312840  0.0
           4924     14.592996  5.132274  0.0

Now the same with a 2D mesh:

>>> df.drop(columns=['z']).plain_mesh.coordinates.head()
                            x         y
element_id node_id
1          1734     14.897208  5.269875
           1582     14.555333  5.355806
           1596     14.630658  4.908741
           4923     14.726271  5.312840
           4924     14.592996  5.132274
'''

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np
import pandas as pd
from pylife import PylifeSignal


@pd.api.extensions.register_dataframe_accessor("plain_mesh")
class PlainMesh(PylifeSignal):
    '''DataFrame accessor to access plain 2D and 3D mesh data, i.e. without connectivity

    Raises
    ------
    AttributeError
        if at least one of the columns `x`, `y` is missing

    Notes
    -----
    The PlainMesh describes meshes whose only geometrical
    information is the coordinates of the nodes or elements. Unlike
    :class:`Mesh` they don't know about connectivity, not even
    about elements and nodes.

    See also
    --------
    :class:`Mesh`: accesses meshes with connectivity information
    :func:`pandas.api.extensions.register_dataframe_accessor()`: concept of DataFrame accessors
    '''
    def _validate(self):
        self._coord_keys = ['x', 'y']
        self.fail_if_key_missing(self._coord_keys)
        if 'z' in self._obj.columns:
            self._coord_keys.append('z')
        self._cached_dimensions = None

    @property
    def dimensions(self):
        """The dimensions of the mesh (2 for 2D and 3 for 3D)

        Note
        ----
        If all the coordinates in z-direction are equal the mesh is considered 2D.
        """
        if self._cached_dimensions is not None:
            return self._cached_dimensions

        if len(self._coord_keys) == 2 or (self._obj.z == self._obj.z.iloc[0]).all():
            self._cached_dimensions = 2
        else:
            self._cached_dimensions = 3

        return self._cached_dimensions

    @property
    def coordinates(self):
        '''Returns the coordinate colums of the accessed DataFrame

        Returns
        -------
        coordinates : pandas.DataFrame
            The coordinates `x`, `y` and if 3D `z` of the accessed mesh
        '''
        return self._obj[self._coord_keys]


@pd.api.extensions.register_dataframe_accessor("mesh")
class Mesh(PlainMesh):

    '''DataFrame accessor to access FEM mesh data (2D and 3D)

    Raises
    ------
    AttributeError
        if at least one of the columns `x`, `y` is missing
    AttributeError
        if the index of the DataFrame is not a two level MultiIndex
        with the names `node_id` and `element_id`

    Notes
    -----
    The Mesh describes how we expect FEM data to look like. It
    consists of nodes identified by `node_id` and elements identified
    by `element_id`. A node playing a role in several elements and an
    element consists of several nodes. So in the DataFrame a `node_id`
    can appear multiple times (for each element, the node is playing a
    role in). Likewise each `element_id` appears multiple times (for
    each node the element consists of).

    The combination `node_id`:`element_id` however, is unique. So the
    table is indexed by a :class:`pandas.MultiIndex` with the level
    names `node_id`, `element_id`.

    See also
    --------
    :class:`PlainMesh`: accesses meshes without connectivity information
    :func:`pandas.api.extensions.register_dataframe_accessor()`: concept of DataFrame accessors

    Examples
    --------
    For an example see :mod:`hotspot`.
    '''
    def _validate(self):
        super()._validate()
        self._cached_element_groups = None
        if set(self._obj.index.names) != set(['element_id', 'node_id']):
            raise AttributeError("A mesh needs a pd.MultiIndex with the names `element_id` and `node_id`")


    @property
    def connectivity(self):
        """The connectivity of the mesh."""
        return self._element_groups['node_id'].apply(np.hstack)

    def vtk_data(self):
        """Make VTK data structure easily plot the mesh with pyVista.

        Returns
        -------
        offsets : ndarray
            An empty numpy array as ``pyVista.UnstructuredGrid()`` still
            demands the argument for the offsets, even though VTK>9 does not
            accept it.
        cells : ndarray
            The location of the cells describing the points in a way
            ``pyVista.UnstructuredGrid()`` needs it
        cell_types : ndarray
            The VTK code for the cell types (see https://github.com/Kitware/VTK/blob/master/Common/DataModel/vtkCellType.h)
        points : ndarray
            The coordinates of the cell points

        Notes
        -----
        This is a convenience function to easily plot a 3D mesh with
        pyVista. It prepares a data structure which can be passed to
        ``pyVista.UnstructuredGrid()``

        Example
        -------
        >>> import pyvista as pv
        >>> grid = pv.UnstructuredGrid(*our_mesh.mesh.vtk_data())
        >>> plotter = pv.Plotter(window_size=[1920, 1080])
        >>> plotter.add_mesh(grid, scalars=our_mesh.groupby('element_id')['val'].mean().to_numpy())
        >>> plotter.show()

        Note the `*` that needs to be added when calling ``pv.UnstructuredGrid()``.
        """
        def choose_element_types_dict():
            return self._element_types_3d if self.dimensions == 3 else self._element_types_2d

        def cells_with_lengths(index, connectivity):
            def locs(nodes):
                return np.array(list(map(index.get_loc, nodes)))

            cells = connectivity.apply(locs)
            return np.array([nd for cell in cells.values for nd in np.insert(cell, 0, cell.shape[0])])

        def calc_cells():
            element_types_dict = choose_element_types_dict()

            groups = self._element_groups['node_id']
            connectivity = groups.apply(np.hstack)
            count = groups.count()

            for total_num, (first_order_num, _) in element_types_dict.items():
                choice = count == total_num
                connectivity[choice] = connectivity[choice].apply(lambda nds: nds[:first_order_num])

            return connectivity, count.apply(lambda c: element_types_dict[c][1]).to_numpy()

        def first_order_points(connectivity):
            points = self._obj.groupby('node_id', sort=True).first()[self._coord_keys]
            nodes = pd.Series([nd for element in connectivity.values for nd in element], name='node_id').unique()
            selection = points.index.isin(nodes)
            return points[selection]

        connectivity, cell_types = calc_cells()
        points = first_order_points(connectivity)
        cells = cells_with_lengths(points.index, connectivity)

        return cells, cell_types, points.to_numpy()

    _element_types_2d = {
        # Resolve number of nodes of element to number of first order nodes and vtk element type
        # see https://kitware.github.io/vtk-examples/site/VTKFileFormats/
        # and https://github.com/Kitware/VTK/blob/master/Common/DataModel/vtkCellType.h
        # number_of_nodes: (number_of_first_order_nodes, vtk_element_type)
        3: (3, 5),  # tri lin
        6: (3, 5),  # tri quad
        4: (4, 9),  # squ lin
        8: (4, 9),  # squ quad
    }
    _element_types_3d = {
        # Resolve number of nodes of element to number of first order nodes and vtk element type
        # see https://kitware.github.io/vtk-examples/site/VTKFileFormats/
        # and https://github.com/Kitware/VTK/blob/master/Common/DataModel/vtkCellType.h
        # number_of_nodes: (number_of_first_order_nodes, vtk_element_type)
        4: (4, 10),   # tet lin
        6: (6, 13),   # wedge lin
        8: (8, 12),   # hex lin
        10: (4, 10),  # tet quad
        15: (6, 26),  # tet wedge
        20: (8, 12),  # hex quad
    }

    @property
    def _element_groups(self):
        if self._cached_element_groups is None:
            self._cached_element_groups = self._obj.reset_index().groupby('element_id')
        return self._cached_element_groups



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import sys
import numpy as np
import odbAccess as ODB
import json


class OdbInterface:

    def __init__(self, odbfile):
        self._odb = ODB.openOdb(odbfile)
        self._asm = self._odb.rootAssembly

        self._index_cache = {}

    def instance_names(self):
        return self._asm.instances.keys()

    def step_names(self):
        return self._odb.steps.keys()

    def frame_names(self, step_name):
        try:
            step = self._odb.steps[step_name]
        except Exception as e:
            return e

        return [frame.frameId for frame in step.frames]

    def nodes(self, instance_name, node_set_name):
        instance = self._instance_or_rootasm(instance_name)

        if node_set_name == b'':
            nodes = instance.nodes
        elif node_set_name in instance.nodeSets.keys():
            nodes = instance.nodeSets[node_set_name].nodes
        elif node_set_name in self._asm.nodeSets.keys():
            nodes = self._asm.nodeSets[node_set_name].nodes[self._asm.instances.keys().index(instance_name)]
        else:
            raise KeyError(node_set_name)

        node_data = np.empty((len(nodes), 3))
        index = np.empty(len(nodes), dtype=np.int32)

        for i, nd in enumerate(nodes):
            index[i] = nd.label
            node_data[i] = nd.coordinates

        return (index, node_data)

    def connectivity(self, instance_name, element_set_name):
        instance = self._instance_or_rootasm(instance_name)

        if element_set_name == b'':
            elements = instance.elements
        elif element_set_name in instance.elementSets.keys():
            elements = instance.elementSets[element_set_name].elements
        elif element_set_name in self._asm.elementSets.keys():
            elements = self._asm.elementSets[element_set_name].elements[self._asm.instances.keys().index(instance_name)]
        else:
            raise KeyError(element_set_name)

        index = np.empty(len(elements), dtype=np.int)
        connectivity = []
        for i, el in enumerate(elements):
            index[i] = el.label
            connectivity.append(list(el.connectivity))

        return (index, connectivity)

    def node_sets(self, instance_name):
        instance = self._instance_or_rootasm(instance_name)
        return instance.nodeSets.keys()

    def element_sets(self, instance_name):
        instance = self._instance_or_rootasm(instance_name)
        return instance.elementSets.keys()

    def node_set(self, instance_name, node_set_name):
        instance = self._instance_or_rootasm(instance_name)

        try:
            node_set = instance.nodeSets[node_set_name]
        except Exception as e:
            return e

        nodes = node_set.nodes[0] if instance_name == b'' else node_set.nodes

        return np.array([node.label for node in nodes], dtype=np.int32)

    def element_set(self, instance_name, element_set_name):
        instance = self._instance_or_rootasm(instance_name)

        try:
            element_set = instance.elementSets[element_set_name]
        except Exception as e:
            return e

        elements = element_set.elements[0] if instance_name == b'' else element_set.elements

        return np.array([element.label for element in elements], dtype=np.int32)

    def variable_names(self, step_name, frame_num):
        try:
            step = self._odb.steps[step_name]
        except Exception as e:
            return e

        try:
            frame = _get_frame(step, frame_num)
        except Exception as e:
            return e

        return frame.fieldOutputs.keys()

    def variable(self, instance_name, step_name, frame_num, variable_name, node_set_name, element_set_name, position=None):

        def block_length(block):
            if block.nodeLabels is not None:
                return block.nodeLabels.shape[0]
            if block.elementLabels is not None:
                return block.elementLabels.shape[0]
            return 0

        def index_block_data(block):
            stack = []
            index_labels = []

            if getattr(block, 'nodeLabels', None) is not None:
                stack.append(block.nodeLabels)
                index_labels.append('node_id')

            if getattr(block, 'elementLabels', None) is not None:
                stack.append(block.elementLabels)
                index_labels.append('element_id')

            if getattr(block, 'integrationPoints', None) is not None:
                stack.append(block.integrationPoints)
                index_labels.append('ipoint_id')

            if getattr(block, 'faces', None) is not None:
                stack.append(block.faces)
                index_labels.append('face_id')

            return np.vstack(stack), index_labels

        try:
            step = self._odb.steps[step_name]
        except Exception as e:
            return e

        try:
            frame = _get_frame(step, frame_num)
        except Exception as e:
            return e

        try:
            field = frame.fieldOutputs[variable_name]
        except Exception as e:
            return e

        instance = self._asm.instances[instance_name]

        region = None
        if node_set_name != b'':
            if node_set_name in instance.nodeSets.keys():
                node_set = instance.nodeSets[node_set_name]
            elif node_set_name in self._asm.nodeSets.keys():
                node_set = self._asm.nodeSets[str(node_set_name)]
            else:
                raise KeyError(node_set_name)
            region = node_set

        if element_set_name != b'':
            if element_set_name in instance.elementSets.keys():
                element_set = instance.elementSets[element_set_name]
            elif element_set_name in self._asm.elementSets.keys():
                element_set = self._asm.elementSets[str(element_set_name)]
            else:
                raise KeyError(element_set_name)
            region = element_set

        position_str = position
        position = _set_position(field, user_request=position_str)
        field = field.getSubset(position=position)

        if region is not None:
            field = field.getSubset(region=region)

        complabels = field.componentLabels if len(field.componentLabels) > 0 else [variable_name]
        blocks = field.bulkDataBlocks

        length = 0
        for block in blocks:
            if block.instance.name != instance_name:
                continue
            length += block_length(block)

        values = np.empty((length, len(complabels)))

        if position in [ODB.INTEGRATION_POINT, ODB.ELEMENT_NODAL, ODB.ELEMENT_FACE]:
            index_dim = 2
        elif position in [ODB.CENTROID, ODB.WHOLE_ELEMENT, ODB.NODAL]:
            index_dim = 1
        index = np.empty((length, index_dim), dtype=np.int32)

        i = 0
        for block in blocks:
            if block.instance.name != instance_name:
                continue

            block_array = block.data
            size = block_array.shape[0]

            index_block, index_labels = index_block_data(block)

            index[i:i+size, :] = index_block.T
            values[i:i+size, :] = block_array
            i += size

        return (complabels, index_labels, index[:i, :], values[:i])

    def _instance_or_rootasm(self, instance_name):
        if instance_name == b'':
            return self._asm
        try:
            return self._asm.instances[instance_name]
        except Exception as e:
            return e

    def history_regions(self, step_name):
        """Get history regions, which belongs to the given step.

        Parameters
        ----------
        step_name : Abaqus steps
            It is always required.

        Returns
        -------
        histRegions : history regions, which belong to the given step. In case of error it gives an error message.
            It is a list of hist regions
        """
        try:

            required_step = self._odb.steps[step_name]
            histRegions = required_step.historyRegions.keys()

            return histRegions

        except Exception as e:
            return e


    def history_outputs(self, step_name, historyregion_name):
        """Get history outputs, which belongs to the given step and history region.

        Parameters
        ----------
        step_name : Abaqus steps
            It is always required.
        historyregion_name: Abaqus history region
            It is always required.

        Returns
        -------
        history_data : history data, which belong to the given step and history region. In case of error it gives an error message.
            It is a list of history outputs.

        """
        try:
           required_step = self._odb.steps[step_name]

           history_data = required_step.historyRegions[historyregion_name].historyOutputs.keys()

           return history_data

        except Exception as e:
            return e


    def history_output_values(self, step_name, historyregion_name, historyoutput_name):
        """Get history output values, which belongs to the given step, history region and history output.

        Parameters
        ----------
        step_name : Abaqus steps
            It is always required.
        historyregion_name: Abaqus history region
            It is always required.
        historyoutput_name: Abaqus history output
            It is always required.

        Returns
        -------
        x : time values of a history output. In case of error it gives an error message.
            It is a list of data.
        y : values of a history output. In case of error it gives an error message.
            It is a list of data.

        """
        try:
           required_step = self._odb.steps[step_name]

           history_data = required_step.historyRegions[historyregion_name].historyOutputs[historyoutput_name].data

           step_time = required_step.totalTime

           xdata = []
           ydata = []
           for ith in history_data:
               xdata.append(ith[0]+step_time)
               ydata.append(ith[1])

           x = np.array(xdata)
           y = np.array(ydata)
           return x,y

        except Exception as e:
            return e

    def history_region_description(self, step_name, historyregion_name):
        """Get history region description, which belongs to the given step and history region.

        Parameters
        ----------
        step_name : Abaqus steps
            It is always required.
        historyregion_name: Abaqus history region
            It is always required.

        Returns
        -------
        history_description : description for history region, which is visible in Abaqus. In case of error it gives an error message.
            It is string.

        """
        try:
           required_step = self._odb.steps[step_name]

           history_description = required_step.historyRegions[historyregion_name].description

           return history_description

        except Exception as e:
            return e


    def history_info(self):
        """Get steps, history regions, history outputs and write into a dictionary.


        Returns
        -------
        A dictionary which contains information about the history of a given odb file.
        In case of error it gives an error message.

        """
        dict1 = {}
        try:
            steps = self._odb.steps.keys()

            for istep in steps:
                regions = self.history_regions(step_name=istep)

                for iregion in regions:

                    idescription = self.history_region_description(step_name= istep, historyregion_name= iregion)

                    outputs = self.history_outputs(step_name= istep, historyregion_name= iregion)

                    for ioutputs in outputs:
                        if "Repeated: key" in ioutputs:
                            outputs.remove(ioutputs)

                    steplist = []
                    for istep2 in steps:
                        try:
                            self._odb.steps[istep2].historyRegions[iregion].description
                            steplist.append(istep2)
                        except Exception as e:
                             continue

                    dict5 = {
                        "History Region" : iregion,
                        "History Outputs" : outputs,
                        "Steps " : steplist
                    }

                    dict1[idescription] = dict(dict5)

            return dict1
        except Exception as e:
            return e


def _set_position(field, user_request=None):
    """Translate string to symbolic constant and define default behavior.

    Parameters
    ----------
    field : Abaqus field object
        Required if ``user_request=None``.
    user_request : string
        Abaqus .inp file terminology (*ELEMENT OUTPUT, position=...).

    Returns
    -------
    position : symbolic constant
        Abaqus Python interface terminology.
    """
    _position_dict = {'INTEGRATION POINTS': ODB.INTEGRATION_POINT,
                      'CENTROIDAL':         ODB.CENTROID,
                      'WHOLE ELEMENT':      ODB.WHOLE_ELEMENT,
                      'NODES':              ODB.ELEMENT_NODAL,
                      'FACES':              ODB.ELEMENT_FACE,
                      'AVERAGED AT NODES':  ODB.NODAL}

    if user_request is not None:
        return _position_dict[user_request]

    odb_pos = field.locations[0].position

    if odb_pos == ODB.INTEGRATION_POINT:
        return ODB.ELEMENT_NODAL

    return odb_pos

def _get_frame(step, frame_id):
    for frame in step.frames:
        if frame_id == frame.frameId:
            return frame

    raise Exception("Invalid frame id %s", frame_id)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import pandas as pd
import numpy as np

import pylife.stress.stresssignal


def test_voigt():
    df = pd.DataFrame({'S11': [1.0], 'S22': [1.0], 'S33': [1.0],
                       'S12': [1.0], 'S13': [1.0], 'S23': [1.0]})
    df.voigt


def test_voigt_fail():
    df = pd.DataFrame({'S11': [1.0], 'S22': [1.0], 'S33': [1.0],
                       'S12': [1.0], 'S31': [1.0], 'S23': [1.0]})
    with pytest.raises(AttributeError, match=r'^StressTensorVoigt.*Missing S13'):
        df.voigt
